from ruamel.yaml import YAML
import os
import re
from mpunet.logging import ScreenLogger
import numpy as np


def _check_deprecated_params(hparams, logger):
    if hparams.get('fit') and hparams["fit"].get("sparse"):
        from mpunet.errors.deprecated_warnings import warn_sparse_param
        warn_sparse_param(logger)


def _check_version(hparams, logger, package="mpunet"):
    from mpunet.bin.version import VersionController
    vc = VersionController(package=package)
    if not vc.check_git():
        logger.warn("Path {} does not contain a Git repository, or Git is not"
                    " installed on this system.\n-- The software verison match "
                    "could not be varified against the hyperparameter file.\n"
                    "-- Software version will not be added to the "
                    "hyperparameter file.".format(vc.git_path))
        return False
    if "__VERSION__" not in hparams:
        e = "Could not infer the software version used to produce the " \
            "hyperparameter file of this project. Using a later " \
            "version of the {} software on this project " \
            "may produce unexpected results. If you wish to continue " \
            "using this software version on this project dir, " \
            "manually add the following line to the hyperparameter file:" \
            " \n\n__VERSION__: {}\n".format(package, vc.version)
        logger.warn(e)
        raise RuntimeWarning(e)
    hp_version = hparams["__VERSION__"]
    if isinstance(hp_version, str) and vc.version != hp_version:
        e = "Parameter file indicates that this project was created " \
            "under {} version {}, but the current " \
            "version is {}. If you wish to continue " \
            "using this software version on this project dir, " \
            "manually add the following line to the hyperparameter " \
            "file:\n\n__VERSION__: {}\n".format(package, hp_version,
                                                vc.version, vc.version)
        logger.warn(e)
        raise RuntimeWarning(e)
    return True


def _set_version(hparams, logger=None, package="mpunet"):
    from mpunet.bin.version import VersionController
    vc = VersionController(package=package)
    if logger:
        vc.log_version(logger)
    v, b, c = vc.version, vc.branch, vc.current_commit
    hparams.set_value(None, "__VERSION__", v, overwrite=True)
    hparams.set_value(None, "__BRANCH__", b, overwrite=True)
    hparams.set_value(None, "__COMMIT__", c, overwrite=True)
    hparams.save_current()


class YAMLHParams(dict):
    def __init__(self, yaml_path, logger=None, no_log=False,
                 no_version_control=False, **kwargs):
        dict.__init__(self, **kwargs)

        # Set logger or default print
        self.logger = logger or ScreenLogger()

        # Set YAML path
        self.yaml_path = os.path.abspath(yaml_path)
        self.string_rep = ""
        self.project_path = os.path.split(self.yaml_path)[0]
        if not os.path.exists(self.yaml_path):
            raise OSError("YAML path '%s' does not exist" % self.yaml_path)
        else:
            with open(self.yaml_path, "r") as yaml_file:
                self.string_rep = yaml_file.read()
            hparams = YAML(typ="safe").load(self.string_rep)

        # Set dict elements
        self.update({k: hparams[k] for k in hparams if k[:4] != "__CB"})

        # Log basic information here...
        self.no_log = no_log
        if not self.no_log:
            self.logger("YAML path:    %s" % self.yaml_path)

        # Version controlling
        _check_deprecated_params(self, self.logger)
        if not no_version_control:
            package = kwargs.get('package') or "mpunet"
            has_git = _check_version(self, self.logger, package)
            if has_git:
                _set_version(self, self.logger if not no_log else None, package)

    @property
    def groups(self):
        groups_re = re.compile(r"\n^(?![ \n])(.*?:.*?\n)", re.MULTILINE)
        start, groups = 0, []
        for iter in re.finditer(groups_re, self.string_rep):
            end = iter.start(0)
            groups.append(self.string_rep[start:end])
            start = end
        groups.append(self.string_rep[start:])
        return groups

    def get_group(self, group_name):
        groups = [g.lstrip("\n").lstrip(" ") for g in self.groups]
        return groups[[g.split(":")[0] for g in groups].index(group_name)]

    def add_group(self, yaml_string):
        group_name = yaml_string.lstrip(" ").lstrip("\n").split(":")[0]

        # Set dict version in memory
        self[group_name] = YAML().load(yaml_string)[group_name]

        # Add pure yaml string to string representation
        self.string_rep += ("\n" + yaml_string.lstrip(" \n") + "\n")

    def delete_group(self, group_name):
        self.string_rep = self.string_rep.replace(self.get_group(group_name), "")
        del self[group_name]

    def get_from_anywhere(self, key, default=None):
        found = []
        for group_str in self:
            group = self[group_str]
            try:
                f = key in group
            except TypeError:
                f = False
            if f:
                found.append((group, group[key]))
        if len(found) > 1:
            self.logger("[ERROR] Found key '%s' in multiple groups (%s)" %
                        (key, [g[0] for g in found]))
        elif len(found) == 0:
            return default
        else:
            return found[0][1]

    def log(self):
        for item in self:
            self.logger("%s\t\t%s" % (item, self[item]))

    def _update_string_line_by_name(self, name, new_value, subdir=None):
        group = self.string_rep if not subdir else self.get_group(subdir)
        lines = group.split("\n")
        found = False
        for i, line in enumerate(lines):
            if found:
                break
            if line.lstrip().startswith(name):
                new_line = line.split(":")[0] + ": {}".format(new_value)
                lines[i] = new_line
                found = True
        if not found:
            raise AttributeError("No field has the name '{}'".format(name))
        new_group = "\n".join(lines)
        self.string_rep = self.string_rep.replace(group, new_group)

    def _set_value_no_subdir(self, name, value, str_value, overwrite,
                             add_if_missing):
        cur_value = self.get(name, None)
        if name in self:
            if cur_value is not None and not overwrite:
                return "Item of name '{}' already set with value '{}'." \
                       " Skipping. (overwrite=False)".format(name, value)
            # Update string representation
            self._update_string_line_by_name(name, str_value)
        elif not add_if_missing:
            raise AttributeError("Entry with name '{}' does not exist and "
                                 "add_if_missing was set to False."
                                 "".format(name))
        else:
            # Add to end of string representation
            self.string_rep = self.string_rep.rstrip("\n") + \
                              "\n\n{}: {}\n".format(name, str_value)
        # Set the value in memory
        self[name] = value

    def _set_value_in_existing_dir(self, subdir, name, value, str_value,
                                   overwrite, add_if_missing):
        cur_value = self[subdir].get(name, None)
        if name in self[subdir]:
            if cur_value is not None and not overwrite:
                return "Entry of name '{}' already set in subdir '{}' "\
                       "with value '{}'. Skipping "\
                       "(overwrite=False).".format(name, subdir, value)
            # Update string representation of new value
            self._update_string_line_by_name(name, str_value, subdir=subdir)
        elif not add_if_missing:
            raise AttributeError("Entry with name '{}' does not exist under "
                                 "subdir '{}' and add_if_missing was set to "
                                 "False.".format(name, subdir))
        else:
            # Add the line to the group
            group = self.get_group(subdir).rstrip(" \n")
            entry = "  {}: {}".format(name, str_value)
            new_group = "{}\n{}".format(group, entry)
            self.string_rep = self.string_rep.replace(group, new_group)
        # Update value in memory
        self[subdir][name] = value

    def _set_value_in_subdir(self, subdir, name, value, str_value, overwrite,
                             add_if_missing):
        if subdir not in self:
            if not add_if_missing:
                raise AttributeError("Subdir '{}' does not "
                                     "exist.".format(subdir))
            else:
                new_group = "{}:\n  {}: {}".format(subdir, name, str_value)
                self.add_group(new_group)
                return "Subdir '{}' does not exist, creating it now... "\
                       "(add_if_missing=True)".format(subdir)
        else:
            return self._set_value_in_existing_dir(
                subdir, name, value, str_value, overwrite, add_if_missing
            )

    def set_value(self, subdir, name, value, overwrite=False, add_if_missing=True):
        # Get propper str rep of value
        if isinstance(value, np.ndarray):
            str_value = np.array2string(value, separator=", ")
        else:
            str_value = str(value)

        if subdir is None:
            status = self._set_value_no_subdir(name, value, str_value,
                                               overwrite=overwrite,
                                               add_if_missing=add_if_missing)
        else:
            status = self._set_value_in_subdir(subdir, name, value, str_value,
                                               overwrite=overwrite,
                                               add_if_missing=add_if_missing)

        # Log what was actually done
        status = status or "Setting value '{}' (type {}) in subdir '{}' " \
                           "with name '{}'".format(str_value, type(value),
                                                   subdir, name)
        self.logger(status)

    def save_current(self, out_path=None):
        # Write to file
        out_path = os.path.abspath(out_path or self.yaml_path)
        if not self.no_log:
            self.logger("Saving current YAML configuration to file:\n", out_path)
        with open(out_path, "w") as out_f:
            out_f.write(self.string_rep)



"""
Mathias Perslev
MSc Bioinformatics

University of Copenhagen
November 2017
"""

import glob
import os
import numpy as np

from pathlib import Path
from .image_pair import ImagePair
from mpunet.logging import ScreenLogger


class ImagePairLoader(object):
    """
    ImagePair data loader object
    Represents a collection of ImagePairs
    """
    def __init__(self,
                 base_dir="./",
                 img_subdir="images",
                 label_subdir="labels",
                 logger=None,
                 sample_weight=1.0,
                 bg_class=0,
                 predict_mode=False,
                 initialize_empty=False,
                 no_log=False,
                 identifier=None,
                 **kwargs):
        """
        Initializes the ImagePairLoader object from all .nii files in a folder
        or pair of folders if labels are also specified.

        If initialize_empty=False, the following actions are taken immediately
        on initialization:
            - All .nii/.nii.gz image files are found in base_dir/img_subdir
            - Unless predict_mode=True, finds all .nii/.nii.gz label files in
              base_dir/label_subdir
            - ImagePair objects are established for all images/image-label
              pairs. Not that since ImagePairs do not eagerly load data,
              the ImagePairLoader also does not immediately load data into mem

        If initialize_empty=True, the class is initialized but no images are
        loaded. Images can be manually added through the add_image and
        add_files methods.

        Args:
            base_dir:           A path to a directory storing the 'img_subdir'
                                and 'label_subdir' sub-folders
            img_subdir:         Name of sub-folder storing .nii images files
            label_subdir:       Name of sub-folder storing .nii labels files
            logger:             mpunet logger object
            sample_weight:      A float giving a global sample weight assigned
                                to all images loaded by the ImagePairLoader
            bg_class            Background class integer to pass to all
                                ImagePair objects. Usually int(0).
            predict_mode:       Boolean whether labels exist for the images.
                                If True, the labels are assumed stored in the
                                label_subdir with names identical to the images
            initialize_empty:   Boolean, if True do not load any images at init
                                This may be useful for manually assigning
                                individual image files to the object.
            no_log:             Boolean, whether to not log to screen/file
            identifier:         Optional name for the dataset
            **kwargs:           Other keywords arguments
        """
        self.logger = logger if logger is not None else ScreenLogger()

        # Set absolute paths to main folder, image folder and label folder
        self.data_dir = Path(base_dir).absolute()
        self.images_path = self.data_dir / img_subdir
        self.identifier = self.data_dir.name

        # Labels included?
        self.predict_mode = predict_mode or not label_subdir
        if not predict_mode:
            self.labels_path = self.data_dir / label_subdir
        else:
            self.labels_path = None

        # Load images unless initialize_empty is specified
        if not initialize_empty:
            # Get paths to all images
            self.image_paths = self.get_image_paths()

            if not predict_mode:
                # Get paths to labels if included
                self.label_paths = self.get_label_paths(img_subdir,
                                                        label_subdir)
            else:
                self.label_paths = None

            # Load all nii objects
            self.images = self.get_image_objects(sample_weight, bg_class)
        else:
            self.images = []

        if not initialize_empty and not self.image_paths:
            raise OSError("No image files found at %s." % self.images_path)
        if not initialize_empty and not predict_mode and not self.label_paths:
            raise OSError("No label files found at %s." % self.labels_path)

        self._id_to_image = self.get_id_to_images_dict()
        if not no_log:
            self._log()

    def __str__(self):
        return "ImagePairLoader(id={}, images={}, data_dir={})".format(
            self.identifier, len(self), self.data_dir
        )

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        return self.images[item]

    def __iter__(self):
        for im in self.images:
            yield im

    def __len__(self):
        return len(self.images)

    def _log(self):
        self.logger(str(self))
        self.logger("--- Image subdir: %s\n--- Label subdir: %s" % (self.images_path,
                                                                    self.labels_path))

    def load(self):
        """ Invokes the 'load' method on all ImagePairs """
        for image in self:
            image.load()

    def unload(self):
        """ Invokes the 'unload' method on all ImagePairs """
        for image in self:
            image.unload()

    @property
    def id_to_image(self):
        """
        Returns:
            A dictionary of image IDs pointing to image objects
        """
        return self._id_to_image

    def get_id_to_images_dict(self):
        return {image.identifier: image for image in self}

    @property
    def n_loaded(self):
        return sum([image.is_loaded for image in self.images])

    def get_by_id(self, image_id):
        """
        Get a specific ImagePair by its string identifier

        Args:
            image_id: String identifier of an ImagePair

        Returns:
            An ImagePair
        """
        return self.id_to_image[image_id]

    def get_random(self, N=1, unique=False):
        """
        Return N random images, with or without re-sampling

        Args:
            N:      Int, number of randomly sampled images to return
            unique: Bool, whether the sampled images should be all unique

        Returns:
            A list of ImagePair objects
        """
        returned = []
        while len(returned) < N:
            if self.queue:
                with self.queue.get() as image:
                    if unique and image in returned:
                        continue
                    else:
                        returned.append(image)
                        yield image
            else:
                image = self.images[np.random.randint(len(self))]
                if unique and image in returned:
                    continue
                else:
                    returned.append(image)
                    yield image

    def _get_paths_from_list_file(self, base_path, fname="LIST_OF_FILES.txt"):
        """
        Loads a set of paths pointing to .nii files in 'base_path'.
        This method is used in the rare cases that images are not directly
        stored in self.images_path or self.labels_path but those paths stores
        a file named 'fname' storing 1 absolute path per line pointing to the
        images to load.

        Args:
            base_path: A path to a folder
            fname:     The filename of the file at 'base_path' that stores the
                       paths to return

        Returns:
            A list of path strings
        """
        # Check if a file listing paths exists instead of actual files at the
        # image sub folder path
        list_file_path = base_path / fname
        images = []
        if os.path.exists(list_file_path):
            with open(list_file_path, "r") as in_f:
                for path in in_f:
                    path = path.strip()
                    if not path:
                        continue
                    images.append(path)
        else:
            raise OSError("File '%s' does not exist. Did you specify "
                          "the correct img_subdir?" % list_file_path)
        return images

    def get_image_paths(self):
        """
        Return a list of paths to all image files in the self.images_path folder

        Returns:
            A list of pathlib.Path
        """
        images = sorted(glob.glob(str(self.images_path / "*.nii*")))
        if not images:
            # Try to load from a file listing paths at the location
            # This is sometimes a format created by the cv_split.py script
            images = self._get_paths_from_list_file(self.images_path)
        return [Path(p) for p in images]

    def get_label_paths(self, img_subdir, label_subdir):
        """
        Return a list of paths to all label files in the self.labels_path folder
        The label paths are assumed to be identical to the image paths with the
        image subdir name replaced by the label subdir name.

        Args:
            img_subdir:   String, name of the image sub-folder
            label_subdir: String, name of the label sub-folder

        Returns:
            A list of pathlib.Path
        """
        if any([img_subdir not in str(p) for p in self.image_paths]):
            raise ValueError("Mismatch between image paths and specified "
                             "img_subdir. The subdir was not found in one or"
                             " more image paths - Do the paths in "
                             "LIST_OF_FILES.txt point to a subdir of name "
                             "'%s'?" % img_subdir)
        return [p.parent.parent / label_subdir / p.name for p in self.image_paths]

    def get_image_objects(self, sample_weight, bg_class):
        """
        Initialize all ImagePair objects from paths at self.image_paths and
        self.label_paths (if labels exist). Note that data is not loaded
        eagerly.

        Args:
            sample_weight: A float giving the weight to assign to the ImagePair
            bg_class:      Background (integer) class

        Returns:
            A list of initialized ImagePairs
        """
        image_objects = []
        if self.predict_mode:
            for img_path in self.image_paths:
                image = ImagePair(img_path,
                                  sample_weight=sample_weight,
                                  bg_class=bg_class,
                                  logger=self.logger)
                image_objects.append(image)
        else:
            for img_path, label_path in zip(self.image_paths, self.label_paths):
                image = ImagePair(img_path, label_path,
                                  sample_weight=sample_weight,
                                  bg_class=bg_class,
                                  logger=self.logger)
                image_objects.append(image)

        return image_objects

    def add_image(self, image_pair):
        """
        Add a single ImagePair object to the ImagePairLoader

        Args:
            image_pair: An ImagePair
        """
        self.images.append(image_pair)
        # Update ID dict
        self._id_to_image = self.get_id_to_images_dict()

    def add_images(self, image_pair_loader):
        """
        Add a set of ImagePair objects to the ImagePairLoader. Input can be
        either a different ImagePairLoader object or a list of ImagePairs.

        Args:
            image_pair_loader: ImagePairLoader or list of ImagePairs

        Returns:
            self
        """
        try:
            self.images += image_pair_loader.images
        except AttributeError:
            # Passed as list?
            self.images += list(image_pair_loader)
        # Update ID dict
        self._id_to_image = self.get_id_to_images_dict()
        return self

    def get_maximum_real_dim(self):
        """
        Returns the longest distance in mm covered by any axis across images
        of this ImagePairLoader.

        Returns:
            A float
        """
        from mpunet.interpolation.sample_grid import get_maximum_real_dim
        return np.max([get_maximum_real_dim(f.image_obj) for f in self])

    def set_scaler_and_bg_values(self, bg_value, scaler, compute_now=False):
        """
        Loads all images and prepares them for iso-live view interpolation
        training by performing the following operations on each:
            1) Loads the image and labels if not already loaded (transparent)
            2) Define proper background value
            3) Setting multi-channel scaler
            4) Setting interpolator object

        Args:
            bg_value:     See ImagePair.set_bg_value
            scaler:       See ImagePair.set_scaler
            compute_now:  TODO
        """
        # Run over volumes: scale, set interpolator, check for affine
        for image in self.id_to_image.values():
            image.set_bg_value(bg_value, compute_now=compute_now)
            image.set_scaler(scaler, compute_now=compute_now)
            image.log_image()



import unittest
import pkgutil
import importlib

_TESTS_PKG = "mpunet.tests"


def get_test_suite(test_class):
    suite = unittest.TestSuite()
    loader = unittest.defaultTestLoader
    suite.addTests(loader.loadTestsFromTestCase(test_class))
    return suite


def _mod_name_to_cls_name(modname):
    """
    Converts the name of a module (such as test_image_pair) to the test suite
    class name (such as TestImagePair)
    """
    return "".join(map(lambda x: x.capitalize(), modname.split("_")))


def _run(test_pkg_name):
    runner = unittest.TextTestRunner()
    test_mod = importlib.import_module("{}.{}".format(_TESTS_PKG,
                                                      test_pkg_name))
    mods = pkgutil.iter_modules(test_mod.__path__)
    s = "[*] Running tests in package '{}' [*]".format(test_pkg_name)
    print("\n" + "-"*len(s) + "\n" + s + "\n" + "-"*len(s))
    suites = []
    for m in mods:
        print("Adding test module '{}'".format(m.name))
        mod = importlib.import_module("{}.{}.{}".format(_TESTS_PKG,
                                                        test_pkg_name, m.name))
        cls_name = _mod_name_to_cls_name(m.name)
        try:
            suite = get_test_suite(getattr(mod, cls_name))
        except AttributeError:
            print("-- Skipping (found no valid test "
                  "suite with name '{}')".format(cls_name))
        else:
            suites.append(suite)
    print("\nRunning...")
    runner.run(unittest.TestSuite(suites))


def run_tests(tests_to_run):
    tests_to_run = tests_to_run.lower()
    from mpunet.tests import get_test_packages
    test_pkgs = get_test_packages()
    if tests_to_run != "all" and tests_to_run not in test_pkgs:
        raise ValueError("Unknown set of tests '{}'. Must be 'all' or one "
                         "of {}.".format(tests_to_run, test_pkgs))
    to_run = test_pkgs if tests_to_run == "all" else [tests_to_run]

    for test_pkg in to_run:
        # Run each package of tests
        _run(test_pkg)


def run_all_tests():
    run_tests("all")


if __name__ == "__main__":
    run_all_tests()



"""
mpunet fusion model training script
Optimizes a mpunet fusion model in a specified project folder

The fusion model takes a set of class-probability vectors, one from each view
of a mpunet model, and merges them into a single class-probability
vector. The fusion model learns to apply additional or reduced weight to each
class as seen in each view of the MultiPlanar model.

The final fusion computes a simple function, but training is relatively compute
intensive as the mpunet model needs to be evaluated over all the
training images in all views. To keep memory load down, the model is optimized
over sub-sets of the training data one-by-one, controlled by the
--images_per_round flag.

Typical usage:
--------------
cd my_project
mp train_fusion
--------------
"""

import random
import numpy as np
import os
from mpunet.image import ImagePairLoader
from mpunet.models import FusionModel
from mpunet.models.model_init import init_model
from mpunet.hyperparameters import YAMLHParams
from mpunet.utils import await_and_set_free_gpu, get_best_model, \
                                  create_folders, highlighted, set_gpu
from mpunet.utils.fusion import predict_and_map, stack_collections
from mpunet.interpolation.sample_grid import get_voxel_grid_real_space
from mpunet.logging import Logger
from mpunet.callbacks import ValDiceScores, PrintLayerWeights
from mpunet.evaluate.metrics import sparse_fg_precision, \
                                             sparse_fg_recall

from tensorflow.keras.optimizers import Adam
from tensorflow.keras.callbacks import CSVLogger, EarlyStopping
from sklearn.utils import shuffle
from argparse import ArgumentParser


def get_argparser():
    parser = ArgumentParser(description='Fit a fusion model for a '
                                        'mpunet project')
    parser.add_argument("--project_dir", type=str, default="./",
                        help='path to mpunet project folder')
    parser.add_argument("--overwrite", action='store_true',
                        help='overwrite previous fusion weights')
    parser.add_argument("--num_GPUs", type=int, default=1,
                        help='Number of GPUs to assign to this job')
    parser.add_argument("--images_per_round", type=int, default=5,
                        help="Number of images to train on in each sub-round."
                             " Larger numbers should be preferred but "
                             "requires potentially large amounts of memory."
                             " Defaults to 5.")
    parser.add_argument("--batch_size", type=int, default=2**17,
                        help="Sets the batch size used in the fusion model "
                             "training process. Lowering this number may "
                             "resolve GPU memory issues. Defaults to 2**17.")
    parser.add_argument("--epochs", type=int, default=30,
                        help="Number of epochs to train on each subset of "
                             "images (default=30)")
    parser.add_argument("--early_stopping", type=int, default=3,
                        help="Number of non-improving epochs before premature "
                             "training round termination is invoked. "
                             "(default=3)")
    parser.add_argument("--continue_training", action='store_true')
    parser.add_argument("--force_GPU", type=str, default="")
    parser.add_argument("--eval_prob", type=float, default=1.0,
                        help="Perform evaluation on only a fraction of the"
                             " computed views (to speed up run-time)")
    parser.add_argument("--wait_for", type=str, default="",
                        help="Waiting for PID to terminate before starting "
                             "training process.")
    parser.add_argument("--dice_weight", type=str, default="uniform")
    return parser


def log(logger, hparams, views, weights, fusion_weights):
    logger("N classes:       %s" % hparams["build"].get("n_classes"))
    logger("Scaler:          %s" % hparams["fit"].get("scaler"))
    logger("Crop:            %s" % hparams["fit"].get("crop_to"))
    logger("Downsample:      %s" % hparams["fit"].get("downsample_to"))
    logger("CF factor:       %s" % hparams["build"].get("complexity_factor"))
    logger("Views:           %s" % views)
    logger("Weights:         %s" % weights)
    logger("Fusion weights:  %s" % fusion_weights)


def contains_all_images(sets, images):
    """
    TODO
    """
    l = [i for s in sets for i in s]
    return all([m in l for m in images])


def make_sets(images, sub_size, N):
    """
    TODO
    """
    sets = []
    for i in range(N):
        sets.append(set(np.random.choice(images, sub_size, replace=False)))
    return sets


def _run_fusion_training(sets, logger, hparams, min_val_images, is_validation,
                         views, n_classes, unet, fusion_model, early_stopping,
                         fm_batch_size, epochs, eval_prob, fusion_weights_path):
    """
    TODO
    """

    for _round, _set in enumerate(sets):
        s = "Set %i/%i:\n%s" % (_round + 1, len(sets), _set)
        logger("\n%s" % highlighted(s))

        # Reload data
        images = ImagePairLoader(**hparams["val_data"])
        if len(images) < min_val_images:
            images.add_images(ImagePairLoader(**hparams["train_data"]))

        # Get list of ImagePair objects to run on
        image_set_dict = {m.identifier: m for m in images if m.identifier in _set}

        # Set scaler and bg values
        images.set_scaler_and_bg_values(
            bg_value=hparams.get_from_anywhere('bg_value'),
            scaler=hparams.get_from_anywhere('scaler'),
            compute_now=False
        )

        # Init LazyQueue and get its sequencer
        from mpunet.sequences.utils import get_sequence
        seq = get_sequence(data_queue=images,
                           is_validation=True,
                           views=views,
                           **hparams["fit"], **hparams["build"])

        # Fetch points from the set images
        points_collection = []
        targets_collection = []
        N_im = len(image_set_dict)
        for num_im, image_id in enumerate(list(image_set_dict.keys())):
            logger("")
            logger(
                highlighted("(%i/%i) Running on %s (%s)" % (num_im + 1, N_im,
                                                            image_id, "val" if
                                                            is_validation[
                                                                image_id] else "train")))

            with seq.image_pair_queue.get_image_by_id(image_id) as image:
                # Get voxel grid in real space
                voxel_grid_real_space = get_voxel_grid_real_space(image)

                # Get array to store predictions across all views
                targets = image.labels.reshape(-1, 1)
                points = np.empty(shape=(len(targets), len(views), n_classes),
                                  dtype=np.float32)
                points.fill(np.nan)

                # Predict on all views
                for k, v in enumerate(views):
                    print("\n%s" % "View: %s" % v)
                    points[:, k, :] = predict_and_map(model=unet,
                                                      seq=seq,
                                                      image=image,
                                                      view=v,
                                                      voxel_grid_real_space=voxel_grid_real_space,
                                                      n_planes='same+20',
                                                      targets=targets,
                                                      eval_prob=eval_prob).reshape(-1, n_classes)

                # add to collections
                points_collection.append(points)
                targets_collection.append(targets)
            print(image.is_loaded)

        # Stack points into one matrix
        logger("Stacking points...")
        X, y = stack_collections(points_collection, targets_collection)

        # Shuffle train
        print("Shuffling points...")
        X, y = shuffle(X, y)

        print("Getting validation set...")
        val_ind = int(0.20*X.shape[0])
        X_val, y_val = X[:val_ind], y[:val_ind]
        X, y = X[val_ind:], y[val_ind:]

        # Prepare dice score callback for validation data
        val_cb = ValDiceScores((X_val, y_val), n_classes, 50000, logger)

        # Callbacks
        cbs = [val_cb,
               CSVLogger(filename="logs/fusion_training.csv",
                         separator=",", append=True),
               PrintLayerWeights(fusion_model.layers[-1], every=1,
                                 first=1000, per_epoch=True, logger=logger)]

        es = EarlyStopping(monitor='val_dice', min_delta=0.0,
                           patience=early_stopping, verbose=1, mode='max')
        cbs.append(es)

        # Start training
        try:
            fusion_model.fit(X, y, batch_size=fm_batch_size,
                             epochs=epochs, callbacks=cbs, verbose=1)
        except KeyboardInterrupt:
            pass
        fusion_model.save_weights(fusion_weights_path)


def entry_func(args=None):

    # Project base path
    args = vars(get_argparser().parse_args(args))
    basedir = os.path.abspath(args["project_dir"])
    overwrite = args["overwrite"]
    continue_training = args["continue_training"]
    eval_prob = args["eval_prob"]
    await_PID = args["wait_for"]
    dice_weight = args["dice_weight"]
    print("Fitting fusion model for project-folder: %s" % basedir)

    # Minimum images in validation set before also using training images
    min_val_images = 15

    # Fusion model training params
    epochs = args['epochs']
    fm_batch_size = args["batch_size"]

    # Early stopping params
    early_stopping = args["early_stopping"]

    # Wait for PID?
    if await_PID:
        from mpunet.utils import await_PIDs
        await_PIDs(await_PID)

    # Fetch GPU(s)
    num_GPUs = args["num_GPUs"]
    force_gpu = args["force_GPU"]
    # Wait for free GPU
    if not force_gpu:
        await_and_set_free_gpu(N=num_GPUs, sleep_seconds=120)
    else:
        set_gpu(force_gpu)

    # Get logger
    logger = Logger(base_path=basedir, active_file="train_fusion",
                    overwrite_existing=overwrite)

    # Get YAML hyperparameters
    hparams = YAMLHParams(os.path.join(basedir, "train_hparams.yaml"))

    # Get some key settings
    n_classes = hparams["build"]["n_classes"]

    if hparams["build"]["out_activation"] == "linear":
        # Trained with logit targets?
        hparams["build"]["out_activation"] = "softmax" if n_classes > 1 else "sigmoid"

    # Get views
    views = np.load("%s/views.npz" % basedir)["arr_0"]
    del hparams["fit"]["views"]

    # Get weights and set fusion (output) path
    weights = get_best_model("%s/model" % basedir)
    weights_name = os.path.splitext(os.path.split(weights)[-1])[0]
    fusion_weights = "%s/model/fusion_weights/" \
                     "%s_fusion_weights.h5" % (basedir, weights_name)
    create_folders(os.path.split(fusion_weights)[0])

    # Log a few things
    log(logger, hparams, views, weights, fusion_weights)

    # Check if exists already...
    if not overwrite and os.path.exists(fusion_weights):
        from sys import exit
        print("\n[*] A fusion weights file already exists at '%s'."
              "\n    Use the --overwrite flag to overwrite." % fusion_weights)
        exit(0)

    # Load validation data
    images = ImagePairLoader(**hparams["val_data"], logger=logger)
    is_validation = {m.identifier: True for m in images}

    # Define random sets of images to train on simul. (cant be all due
    # to memory constraints)
    image_IDs = [m.identifier for m in images]

    if len(images) < min_val_images:
        # Pick N random training images
        diff = min_val_images - len(images)
        logger("Adding %i training images to set" % diff)

        # Load the training data and pick diff images
        train = ImagePairLoader(**hparams["train_data"], logger=logger)
        indx = np.random.choice(np.arange(len(train)), diff, replace=diff > len(train))

        # Add the images to the image set set
        train_add = [train[i] for i in indx]
        for m in train_add:
            is_validation[m.identifier] = False
            image_IDs.append(m.identifier)
        images.add_images(train_add)

    # Append to length % sub_size == 0
    sub_size = args["images_per_round"]
    rest = int(sub_size*np.ceil(len(image_IDs)/sub_size)) - len(image_IDs)
    if rest:
        image_IDs += list(np.random.choice(image_IDs, rest, replace=False))

    # Shuffle and split
    random.shuffle(image_IDs)
    sets = [set(s) for s in np.array_split(image_IDs, len(image_IDs)/sub_size)]
    assert(contains_all_images(sets, image_IDs))

    # Define fusion model (named 'org' to store reference to orgiginal model if
    # multi gpu model is created below)
    fusion_model = FusionModel(n_inputs=len(views), n_classes=n_classes,
                               weight=dice_weight,
                               logger=logger, verbose=False)

    if continue_training:
        fusion_model.load_weights(fusion_weights)
        print("\n[OBS] CONTINUED TRAINING FROM:\n", fusion_weights)

    import tensorflow as tf
    with tf.distribute.MirroredStrategy().scope():
        # Define model
        unet = init_model(hparams["build"], logger)
        print("\n[*] Loading weights: %s\n" % weights)
        unet.load_weights(weights, by_name=True)

    # Compile the model
    logger("Compiling...")
    metrics = ["sparse_categorical_accuracy", sparse_fg_precision, sparse_fg_recall]
    fusion_model.compile(optimizer=Adam(lr=1e-3),
                         loss=fusion_model.loss,
                         metrics=metrics)
    fusion_model._log()

    try:
        _run_fusion_training(sets, logger, hparams, min_val_images,
                             is_validation, views, n_classes, unet,
                             fusion_model, early_stopping, fm_batch_size,
                             epochs, eval_prob, fusion_weights)
    except KeyboardInterrupt:
        pass
    finally:
        if not os.path.exists(os.path.split(fusion_weights)[0]):
            os.mkdir(os.path.split(fusion_weights)[0])
        # Save fusion model weights
        # OBS: Must be original model if multi-gpu is performed!
        fusion_model.save_weights(fusion_weights)


if __name__ == "__main__":
    entry_func()



import pickle
import sqlite3


class DefaultTablesTable(object):

    def __init__(self):
        self.table_name = self.__class__.__name__

    def get_create_query(self, **kwargs):
        return f"CREATE TABLE {self.table_name} (id INTEGER PRIMARY KEY," \
               f"table_name VARCHAR(255),kwargs BLOB);"

    def get_insert_query(self, table_name, **kwargs):
        bdata = pickle.dumps(kwargs, pickle.HIGHEST_PROTOCOL)
        return f"INSERT INTO {self.table_name} (table_name,kwargs) VALUES " \
               f"({table_name}, {sqlite3.Binary(bdata)});"


class ResultsByView(object):
    def __init__(self, n_classes):
        self.n_classes = n_classes
        self.table_name = self.__class__.__name__

    def get_create_query(self, **kwargs):
        query = f"CREATE TABLE {self.table_name} (id INTEGER PRIMARY KEY," \
                f"image_id VARCHAR(255) NOT NULL,mean_dice FLOAT(5),"

        for i in range(self.n_classes):
            query += f"class_{i}_dice FLOAT(5),"
        return query[:-1] + ");"

    def get_insert_query(self, image_id, mean_dice, per_class_dices):
        query = f"INSERT INTO {self.table_name} (image_id,mean_dice,"
        for i in range(self.n_classes):
            query += f"class_{i}_dice,"
        query = query[:-1]
        query += f") VALUES ('{image_id}',{mean_dice},"
        for i in range(self.n_classes):
            query += f"{per_class_dices[i]},"
        return query[:-1] + ");"



import unittest


class TestSystem(unittest.TestCase):

    minimum_mem_gib = 8.0
    minimum_cpu_cores = 4
    minimum_gpus = 1

    def test_enough_memory(self):
        """ Tests if the system has the recommended amount of memory. """
        import psutil
        mem = psutil.virtual_memory()
        mem_total_gib = mem.total * (2**-30)
        self.assertGreaterEqual(mem_total_gib, self.minimum_mem_gib,
                                msg="The total available memory of this system"
                                    " is {} GiB. For most practical "
                                    "application more than {} GiB will be "
                                    "needed to run mpunet optimally. "
                                    "For testing purposes, you may disregard "
                                    "this error.".format(mem_total_gib,
                                                         self.minimum_mem_gib))

    def test_enough_cpu_cores(self):
        """ Tests if the system has the recommended number of CPU cores. """
        import psutil
        cpus = psutil.cpu_count(logical=True)
        self.assertGreaterEqual(cpus, self.minimum_cpu_cores,
                                msg="The number of logical CPU cores of this "
                                    "system is {}. We recommend at least {} "
                                    "for mpunet to run optimally. "
                                    "For testing purposes, you may disregard "
                                    "this error.".format(cpus,
                                                         self.minimum_cpu_cores))

    def test_nvidia_smi_available(self):
        """ Tests if the nvidia-smi is available """
        msg = "nvidia-smi does not seem to be available - " \
              "this could indicate that the nvidia drivers are not installed."
        import shutil
        import os
        path = shutil.which("nvidia-smi")
        self.assertIsNot(path, None, msg)
        self.assertTrue(os.path.exists(path), msg)

    def test_gpu_count(self):
        """ Tests if the system has at least 1 nvidia GPU """
        from subprocess import check_output
        try:
            gpu_list = check_output("nvidia-smi "
                                    "--query-gpu=gpu_name "
                                    "--format=csv".split(" "),
                                    universal_newlines=True).split("\n")[1:]
        except FileNotFoundError as e:
            raise FileNotFoundError("nvidia-smi does not seem to be "
                                    "installed - this could indicate missing "
                                    "GPU drivers.") from e
        n_gpus = len(list(filter(None, gpu_list)))
        self.assertGreaterEqual(n_gpus, self.minimum_gpus,
                                msg="The number of GPUs on this "
                                    "system is {}. We recommend at least {} "
                                    "for mpunet to run optimally. "
                                    "For testing purposes, you may disregard "
                                    "this error.".format(n_gpus,
                                                         self.minimum_gpus))



from mpunet.sequences import BaseSequence
from mpunet.logging import ScreenLogger
import numpy as np


class IsotrophicLiveViewSequence(BaseSequence):
    def __init__(self, image_pair_queue, dim, batch_size, n_classes,
                 real_space_span=None, noise_sd=0., force_all_fg="auto",
                 fg_batch_fraction=0.50, label_crop=None, logger=None,
                 is_validation=False, list_of_augmenters=None, flatten_y=False,
                 **kwargs):
        super().__init__()

        # Validation or training batch generator?
        self.is_validation = is_validation

        # Set logger or default print
        self.logger = logger or ScreenLogger()

        # Set views and attributes for plane sample generation
        self.sample_dim = dim
        self.real_space_span = real_space_span
        self.noise_sd = noise_sd if not self.is_validation else 0.

        # Set data
        self.image_pair_queue = image_pair_queue

        # Augmenter, applied to batch at creation time
        # Do not augment validation data
        self.list_of_augmenters = list_of_augmenters if not self.is_validation else None

        # Batch creation options
        self.batch_size = batch_size
        self.n_classes = n_classes
        self.flatten_y = flatten_y

        # Minimum fraction of slices in each batch with FG
        self.force_all_fg_switch = force_all_fg
        self.fg_batch_fraction = fg_batch_fraction

        # Foreground label settings
        self.fg_classes = np.arange(1, self.n_classes)
        if self.fg_classes.shape[0] == 0:
            self.fg_classes = [1]

        # Set potential label label_crop
        self.label_crop = np.array([[0, 0], [0, 0]]) if label_crop is None else label_crop

    def __len__(self):
        """ Undefined, return some high number - a number is need in keras """
        return int(10**12)

    def __getitem__(self, idx):
        raise NotImplemented

    @property
    def n_samples(self):
        return len(self)

    @property
    def batch_size(self):
        return self._batch_size

    @batch_size.setter
    def batch_size(self, value):
        if value < 0:
            raise ValueError("Invalid batch size of %i" % value)
        self._batch_size = value

    @property
    def n_fg_slices(self):
        return int(np.ceil(self.batch_size * self.fg_batch_fraction))

    @property
    def force_all_fg(self):
        if isinstance(self.force_all_fg_switch, str) and \
                self.force_all_fg_switch.lower() == "auto":
            # If the batch size number is bigger than the total number of
            # FG classes, we force the batch to contain at least 1 voxel of
            # each class in one of the slices. If the total number of classes
            # exceeds the batch size, it may not be possible to have all
            # classes simultaneously. Can be overwritten with False/True.
            return self.batch_size > len(self.fg_classes)
        else:
            return self.force_all_fg_switch

    def _crop_labels(self, batch_y):
        return batch_y[:, self.label_crop[0, 0]:-self.label_crop[0, 1],
                       :self.label_crop[1, 0]:-self.label_crop[1, 1]]

    def is_valid_im(self, im, bg_value):
        # Image slice should not be out of bounds (complete background)
        valid = []
        for i, chn_bg_val in enumerate(bg_value):
            valid.append(np.any(~np.isclose(im[..., i], chn_bg_val)))
        return any(valid)

    def validate_lab_vec(self, lab, has_fg, cur_batch_size):
        new_mask = has_fg + np.isin(self.fg_classes, lab)
        if np.all(new_mask):
            return True, new_mask
        elif np.sum(new_mask == 0) < (self.batch_size - cur_batch_size):
            # No FG, but there are still enough random slices left to fill the
            # minimum requirement
            return True, new_mask
        else:
            # No FG, but there is not enough random slices left to fill the
            # minimum requirement. Discard the slice and sample again.
            return False, has_fg

    def validate_lab(self, lab, has_fg, cur_batch_size, debug=False):
        valid = np.any(np.isin(self.fg_classes, lab))

        if debug:
            print(valid, self.fg_classes)
            print(np.unique(lab, return_counts=True))
            print(self.batch_size, cur_batch_size, self.n_fg_slices, has_fg)

        if valid:
            return True, 1
        elif (self.n_fg_slices - has_fg) < (self.batch_size - cur_batch_size):
            # No FG, but there are still enough random slices left to fill the
            # minimum requirement
            return True, 0
        else:
            # No FG, but there is not enough random slices left to fill the
            # minimum requirement. Discard the slice and sample again.
            return False, 0

    def augment(self, batch_x, batch_y, batch_w, bg_values):
        # Apply further augmentation?
        if self.list_of_augmenters:
            for aug in self.list_of_augmenters:
                batch_x, batch_y, batch_w = aug(batch_x=batch_x,
                                                batch_y=batch_y,
                                                batch_w=batch_w,
                                                bg_values=bg_values)

        return batch_x, batch_y, batch_w

    def scale(self, batch_x, scalers):
        return [scaler.transform(im) for im, scaler in zip(batch_x, scalers)]

    def prepare_batches(self, batch_x, batch_y, batch_w):
        # Crop labels if necessary
        if self.label_crop.sum() != 0:
            batch_y = self._crop_labels(batch_y)

        # Reshape X and y
        batch_x = np.asarray(batch_x)
        batch_y = np.asarray(batch_y)
        batch_w = np.asarray(batch_w)

        if self.flatten_y:
            # Flatten labels in spatial dimensions; needed for sample weighting
            # to work correctly in newer versions of TF.
            # Note: model output must be flattened likewise!
            batch_y = batch_y.reshape((len(batch_y), -1, 1))
        elif batch_y.shape[-1] != 1:
            batch_y = np.asarray(batch_y).reshape(batch_y.shape + (1,))

        return batch_x, batch_y, batch_w



import os
import numpy as np
import tensorflow as tf
from mpunet.callbacks import init_callback_objects
from mpunet.logging import ScreenLogger
from mpunet.callbacks import (SavePredictionImages, Validation,
                              FGBatchBalancer, DividerLine,
                              LearningCurve, MemoryConsumption,
                              MeanReduceLogArrays, remove_validation_callbacks)
from mpunet.utils import ensure_list_or_tuple
from mpunet.train.utils import (ensure_sparse,
                                init_losses,
                                init_metrics,
                                init_optimizer)
from tensorflow.python.framework.errors_impl import (ResourceExhaustedError,
                                                     InternalError)


def get_steps(sequence, im_per_epoch=None):
    """ Returns the number of gradient steps to take in an epoch """
    if im_per_epoch:
        steps = int(np.ceil(im_per_epoch / sequence.batch_size))
    else:
        steps = len(sequence)
    return steps


class Trainer(object):
    """
    Handles initialization and logging of model fitting sessions.
    """
    def __init__(self, model, logger=None):
        """
        Init. simply accepts a model and stores it.
        Optionally, an 'org_model' (original model) may be passed and stored
        as well. This is for training multi-GPU models prepared by the
        tf.keras.utils.multi_gpu_model utility, which returns a new, split
        model for training (passed as 'model' parameter here). For properly
        saving the model parameter, however, it is recommended to use the
        original, non-split model (here passed as 'org_model').

        Args:
            model:      (tf.keras Model) Initialized model to train
            org_model:  (tf.keras Model) Optional single-GPU version for the
                                         passed 'model' parameter.
            logger:     (Logger)         Optional Logger instance
        """
        self.model = model
        self.logger = logger if logger is not None else ScreenLogger()

    def compile_model(self, optimizer, loss, metrics, reduction,
                      check_sparse=False, optimizer_kwargs={}, loss_kwargs={},
                      **kwargs):
        """
        Compile the stored tf.keras Model instance stored in self.model
        Sets the loss function, optimizer and metrics

        Args:
            optimizer:        (string) The name of a tf.keras.optimizers Optimizer
            optimizer_kwargs: (dict)   Key-word arguments passed to the Optimizer
            loss:             (string) The name of a tf.keras.losses or
                                       MultiPlanarUnet loss function
            metrics:          (list)   List of tf.keras.metrics or
                                       mpunet metrics.
            reduction:        TODO
            check_sparse:     TODO
            **kwargs:         (dict)   Key-word arguments passed to losses
                                       and/or metrics that accept such.
        """
        # Make sure sparse metrics and loss are specified as sparse
        metrics = ensure_list_or_tuple(metrics)
        losses = ensure_list_or_tuple(loss)
        if check_sparse:
            ensure_sparse(metrics+losses)

        # Initialize optimizer, loss(es) and metric(s) from tf.keras or
        # mpunet
        optimizer = init_optimizer(optimizer, self.logger, **optimizer_kwargs)
        losses = init_losses(losses, self.logger, **kwargs)
        for i, loss in enumerate(losses):
            try:
                losses[i] = loss(reduction=reduction, **loss_kwargs)
            except (ValueError, TypeError):
                raise TypeError("All loss functions must currently be "
                                "callable and accept the 'reduction' "
                                "parameter specifying a "
                                "tf.keras.losses.Reduction type. If you "
                                "specified a keras loss function such as "
                                "'sparse_categorical_crossentropy', change "
                                "this to its corresponding loss class "
                                "'SparseCategoricalCrossentropy'. If "
                                "you implemented a custom loss function, "
                                "please raise an issue on GitHub.")
        metrics = init_metrics(metrics, self.logger, **kwargs)

        # Compile the model
        self.model.compile(optimizer=optimizer, loss=losses, metrics=metrics)
        self.logger("Optimizer:   %s" % optimizer)
        self.logger("Loss funcs:  %s" % losses)
        self.logger("Metrics:     %s" % init_metrics)
        return self

    def fit(self, train, val, batch_size, no_im=False, **fit_kwargs):
        """
        Fit the stored tf.keras Model (self.model) on a set of data.

        The 'fit' method is a wrapper around the hidden '_fit' method. It
        handles KeyboardInterrupts (--> stopping training prematurely), TF
        GPU memory errors (--> batch_size is reduced by 2 and training
        restarted), and other exceptions (--> error logged and training
        terminated).

        Please refer to the self._fit method for 'fit_kwargs' argument details.

        Args:
            train:      TODO
            val:        TODO
            batch_size: (int)  The initial batch size to run training with
            no_im:      TODO
            fit_kwargs: (dict) Keyword arguments passed to self._fit
        """
        # Crop labels?
        if hasattr(self.model, "label_crop"):
            train.label_crop = self.model.label_crop
            if val is not None:
                val.label_crop = self.model.label_crop
        if type(train).__name__ == "MultiTaskSequence":
            self.logger("-- Skipping saving images (not yet implemented for"
                        " MultiTaskSequences).")
            no_im = True
        # Save a few images to disk for inspection
        if no_im:
            self.logger("No images saved (--no_images flag is set)")
        else:
            from mpunet.utils.plotting import save_images
            im_path = os.path.join(self.logger.base_path, "images")
            save_images(train, val, im_path, self.logger)

        # Start fitting
        fitting = True
        while fitting:
            try:
                self._fit(train=train,
                          val=val,
                          batch_size=batch_size,
                          no_im=no_im,
                          **fit_kwargs)
                fitting = False
            except (ResourceExhaustedError, InternalError):
                # Reduce batch size
                batch_size -= 2
                self.logger("\n\n[MEMORY ERROR] Reducing batch size "
                            "by 2 (now %i)" % batch_size)
                if batch_size < 1:
                    self.logger("[ERROR] Batch size negative or zero!")
                    fitting = False
            except KeyboardInterrupt:
                fitting = False
            except Exception as e:
                self.logger(e)
                raise e

        try:
            if train.image_pair_loader.queue:
                train.image_pair_loader.queue.stop()
            if val.image_pair_loader.queue:
                val.image_pair_loader.queue.stop()
        except AttributeError:
            # Multi-tasking, train.image_pair_loader will be a list
            # TODO: Make all sequences store a reference to the queue
            pass

        self.logger("Training stopped.")
        self.logger.print_calling_method = True
        return self.model

    def _fit(self,
             train,
             val,
             batch_size,
             n_epochs,
             callbacks,
             train_im_per_epoch,
             val_im_per_epoch,
             val_ignore_class_zero=True,
             no_im=False,
             verbose=1,
             init_epoch=0,
             use_multiprocessing=False,
             **unused):
        train.batch_size = batch_size

        # Get number of steps per train epoch
        train_steps = get_steps(train, train_im_per_epoch)
        self.logger("Using %i steps per train epoch (total batches=%i)" %
                    (train_steps, len(train)))

        if val is None:
            # No validation to be performed, remove callbacks that might need
            # validation data to function properly
            remove_validation_callbacks(callbacks, self.logger)
        else:
            val.batch_size = batch_size
            val_steps = get_steps(val, val_im_per_epoch)
            self.logger("Using %i steps per val epoch (total batches=%i)" %
                        (val_steps, len(val)))
            # Add validation callback
            # Important: Should be first in callbacks list as other CBs may
            # depend on the validation metrics/loss
            validation = Validation(val,
                                    steps=val_steps,
                                    ignore_class_zero=val_ignore_class_zero,
                                    logger=self.logger,
                                    verbose=verbose)
            callbacks = [validation] + callbacks

        # Add various callbacks for plotting learning curves etc.
        # Get FGBatchBalancer callbacks, etc.
        if hasattr(train, "n_fg_slices"):
            callbacks.append(FGBatchBalancer(train, logger=self.logger))
        if not no_im:
            # Add save images cb
            callbacks.append(SavePredictionImages(train, val))
        callbacks.insert(1, MeanReduceLogArrays())
        # callbacks.insert(1, MemoryConsumption(logger=self.logger))
        callbacks.append(LearningCurve(logger=self.logger))
        callbacks.append(DividerLine(self.logger))

        # Get initialized callback objects
        callbacks, cb_dict = init_callback_objects(callbacks, self.logger)

        # If ModelCheckPointClean is used, set the original model to store
        # the correct weights when using multi-GPU models
        cb = cb_dict.get("ModelCheckPointClean")
        if cb:
            cb.org_model = self.model  # TEMP TODO

        # Init TF dataset with DATA autosharding
        dtypes, shapes = list(zip(*map(lambda x: (x.dtype, x.shape), train[0])))
        train = tf.data.Dataset.from_generator(train, dtypes, shapes)

        # Fit the model
        # is_queued = bool(train.image_pair_loader.queue)
        self.logger.active_log_file = "training"
        self.logger.print_calling_method = False
        self.model.fit(
            train,
            steps_per_epoch=train_steps,
            epochs=n_epochs,
            callbacks=callbacks,
            initial_epoch=init_epoch,
            use_multiprocessing=use_multiprocessing,
            workers=5,
            max_queue_size=5,
            shuffle=False,  # Determined by the chosen Sequence class
            verbose=verbose
        )



from mpunet.logging import ScreenLogger
from mpunet.image.auditor import Auditor
from mpunet.image import ImagePairLoader
from mpunet.sequences.utils import get_sequence
import numpy as np
import os


"""
A collection of functions that prepares data for feeding to various models in
the mpunet.models packages. All functions should follow the following
specification:

f(hparams, just_one, no_val, logger, mtype, base_path), 
--> return train, val, hparams
See function docstrings for further description 

and return an object of type keras.Sequence feeding valid train and validation
inputs to the model. Should also return the hparams object for clarity.
"""


def _base_loader_func(hparams, just_one, no_val, logger, mtype):
    """
    Base loader function used for all models. This function performs a series
    of actions:

    1) Loads train, val and test data according to hparams
    2) Performs a hparam audit on the training + validation images
    3) If any audited parameters were not manually specified, updates the
       hparams dict with the audited values and updates the YAML file on disk
    4) If just_one, discards all but the first training and validation images
    5) Initializes a ImageQueue object on the training and validation data
       if needed.

    Args:
        hparams:   A mpunet.train.YAMLHParams object
        just_one:  A bool specifying whether to keep only the first train and
                   validation samples (for quick testing purposes)
        no_val:    A bool specifying whether to omit validation data entirely
                   Note: This setting applies even if validation data is
                   specified in the YAMLHparams object
        logger:    A mpunet.logger object
        mtype:     A string identifier for the dimensionality of the model,
                   currently either '2d', '3d'
                   (upper/lower ignored)

    Returns:
        train_data: An ImagePairLoader object storing the training images
        val_data:   An ImagePairLoader object storing the validation images, or
                    an 'empty' ImagePairLoader storing no images if no_val=True
        logger:     The passed logger object or a ScreenLogger object
        auditor:    An auditor object storing statistics on the training data
    """

    # Get basic ScreenLogger if no logger is passed
    logger = logger or ScreenLogger()
    logger("Looking for images...")

    # Get data loaders
    train_data = ImagePairLoader(logger=logger, **hparams["train_data"])
    val_data = ImagePairLoader(logger=logger, **hparams["val_data"])

    # Audit
    lab_paths = train_data.label_paths + val_data.label_paths
    auditor = Auditor(train_data.image_paths + val_data.image_paths,
                      nii_lab_paths=lab_paths, logger=logger,
                      dim_3d=hparams.get_from_anywhere("dim") or 64,
                      hparams=hparams)

    # Fill hparams with audited values, if not specified manually
    auditor.fill(hparams, mtype)

    # Add augmented data?
    if hparams.get("aug_data"):
        aug_data = hparams["aug_data"]
        if "include" not in aug_data:
            logger.warn("Found 'aug_data' group, but the group does not "
                        "contain the key 'include', which is required in "
                        "version 2.0 and above. OBS: Not including aug data!")
        elif aug_data["include"]:
            logger("\n[*] Adding augmented data with weight ", aug_data["sample_weight"])
            train_data.add_images(ImagePairLoader(logger=logger, **aug_data))

    if just_one:
        # For testing purposes, run only on one train and one val image?
        logger("[**NOTTICE**] Only running on first train & val samples.")
        train_data.images = [train_data.images[0]]
        val_data.images = [val_data.images[0]]
    if no_val:
        # Run without performing validation (even if specified in param file)
        val_data.images = []

    # Set per-image properties
    for dataset in (train_data, val_data):
        logger("Preparing dataset {}".format(dataset))
        dataset.set_scaler_and_bg_values(bg_value=hparams.get_from_anywhere('bg_value'),
                                         scaler=hparams.get_from_anywhere('scaler'),
                                         compute_now=False)

    from mpunet.image.queue.utils import get_data_queues
    max_loaded = hparams['fit'].get('max_loaded')
    train_queue, val_queue = get_data_queues(
        train_dataset=train_data,
        val_dataset=val_data,
        train_queue_type='limitation' if max_loaded else "eager",
        val_queue_type='eager',
        max_loaded=max_loaded,
        num_access_before_reload=hparams['fit'].get('num_access'),
        logger=logger
    )

    return train_queue, val_queue, logger, auditor


def load_or_create_views(hparams, continue_training, logger, base_path, auditor):
    """
    TODO

    Args:

    """
    views = hparams["fit"]["views"]
    if not continue_training:
        if isinstance(views, int):
            from mpunet.interpolation.sample_grid import sample_random_views_with_angle_restriction
            views = sample_random_views_with_angle_restriction(views, 60,
                                                               auditor=auditor,
                                                               logger=logger)
            hparams["fit"]["views"] = views
        elif isinstance(views, (list, tuple)):
            if not hparams["fit"]["intrp_style"] == "iso_live":
                logger("[Note] Pre-adding noise to views (SD: %s)" %
                       hparams["fit"]["noise_sd"])
                # Apply noise to views
                from mpunet.utils import add_noise_to_views
                hparams["fit"]["views"] = add_noise_to_views(hparams["fit"]["views"],
                                                             hparams["fit"]["noise_sd"])
                hparams["fit"]["noise_sd"] = False
        else:
            raise ValueError("Invalid 'views' input '%s'. Must be list or "
                             "single integer" % views)
        logger("View SD:     %s" % hparams["fit"].get("noise_sd"))

        # Save views
        np.savez(os.path.join(base_path, "views"), hparams["fit"]["views"])

        # Plot views
        from mpunet.utils.plotting import plot_views
        plot_views(views, os.path.join(base_path, "views.png"))
    else:
        # Fetch views from last session
        view_path = os.path.join(base_path, "views.npz")
        hparams["fit"]["views"] = np.load(view_path)["arr_0"]


def get_sequencers(train_queue, val_queue, logger, hparams):
    """
    TODO

    Args:

    """
    # Get Sequence generators
    logger("Preparing sequence objects...")
    sequencers = []
    for queue, is_val in ((train_queue, False), (val_queue, True)):
        if not queue:
            sequencers.append(None)
        else:
            sequencers.append(get_sequence(
                data_queue=queue,
                is_validation=is_val,
                logger=logger,
                dim=hparams["build"]["dim"],
                n_classes=hparams["build"]["n_classes"],
                **hparams["fit"]
            ))
    train, val = sequencers
    return train, val


def prepare_for_multi_view_unet(hparams, just_one=False, no_val=False,
                                continue_training=False, logger=None,
                                base_path='./'):
    """
    TODO

    Args:

    """
    # Load the data
    train_queue, val_queue, logger, auditor = _base_loader_func(hparams, just_one,
                                                                no_val, logger, "2d")

    # Load or create a set of views (determined by 'continue_training')
    # This function will add the views to hparams["fit"]["views"] and
    # store the views on disk at base_path/views.npz.
    load_or_create_views(hparams=hparams,
                         continue_training=continue_training,
                         logger=logger,
                         base_path=base_path,
                         auditor=auditor)

    # Print views in use
    logger("Views:       N=%i" % len(hparams["fit"]["views"]))
    logger("             %s" % ((" " * 13).join([str(v) + "\n" for v in hparams["fit"]["views"]])))

    # Get Sequence generators
    train, val = get_sequencers(train_queue, val_queue, logger, hparams)
    return train, val, hparams


def prepare_for_3d_unet(hparams, just_one=False, no_val=False, logger=None,
                        continue_training=None, base_path="./"):
    """
    TODO

    Args:

    """
    # Load the data
    train_queue, val_queue, logger, auditor = _base_loader_func(hparams, just_one,
                                                                no_val, logger, "3d")
    train, val = get_sequencers(train_queue, val_queue, logger, hparams)
    return train, val, hparams


def prepare_for_multi_task_2d(hparams, just_one=False, no_val=False, logger=None,
                              continue_training=None, base_path="./"):
    """
    TODO

    Args:

    """
    raise NotImplementedError(
        'Multi task training is not yet implemented for mpunet>=0.2.6. '
        'Please raise an issue on GitHUb.'
    )
    from mpunet.hyperparameters import YAMLHParams
    # Get image loaders for all tasks
    tasks = []
    for name, task_hparam_file in zip(*hparams["tasks"].values()):
        task_hparams = YAMLHParams(task_hparam_file)
        type_ = 'multi_task_2d'
        train_data, val_data, logger, auditor = _base_loader_func(task_hparams,
                                                                  just_one,
                                                                  no_val,
                                                                  logger,
                                                                  mtype=type_)

        task = {
            "name": name,
            "hparams": task_hparams,
            "train": train_data,
            "val": val_data
        }
        tasks.append(task)

    # Set various build hparams
    fetch = ("n_classes", "dim", "n_channels",
             "out_activation", "biased_output_layer")
    field = "task_specifics"
    for f in fetch:
        hparams["build"][f] = tuple([t["hparams"][field][f] for t in tasks])

    # Add task names to build dir
    hparams["build"]["task_names"] = hparams["tasks"]["task_names"]

    # Load or create a set of views (determined by 'continue_training')
    # This function will add the views to hparams["fit"]["views"] and
    # store the views on disk at base_path/views.npz.
    load_or_create_views(hparams=hparams,
                         continue_training=continue_training,
                         logger=logger,
                         base_path=base_path,
                         auditor=None)

    # Get per-task sequences
    train_seqs = []
    val_seqs = []
    for task in tasks:
        logger("Fetching sequences for task %s" % task["name"])

        # Create hparams dict that combines the common hparams and
        # task-specific hparams
        task_hparams = dict(hparams["fit"])
        task_hparams.update(task["hparams"]["task_specifics"])

        # Get sequences for training and validation
        train = task["train"].get_sequencer(is_validation=False, **task_hparams)
        val = task["val"].get_sequencer(is_validation=True, **task_hparams)

        # Add to lists
        train_seqs.append(train)
        val_seqs.append(val)

    # Create the training and validation sequences
    # These will produce batches shared across the N tasks
    from mpunet.sequences import MultiTaskSequence
    train = MultiTaskSequence(train_seqs, hparams["build"]["task_names"])
    val = MultiTaskSequence(val_seqs, hparams["build"]["task_names"])

    return train, val, hparams



import os
import argparse
import subprocess

from multiprocessing import Process, Lock, Queue, Event
from mpunet.utils import create_folders
from mpunet.bin.init_project import copy_yaml_and_set_data_dirs
from mpunet.logging import Logger


def get_parser():
    parser = argparse.ArgumentParser(description="Prepare a data folder for a"
                                                 "CV experiment setup.")
    parser.add_argument("--CV_dir", type=str, required=True,
                        help="Directory storing split subfolders as output by"
                             " cv_split.py")
    parser.add_argument("--out_dir", type=str, default="./splits",
                        help="Folder in which experiments will be run and "
                             "results stored.")
    parser.add_argument("--num_GPUs", type=int, default=1,
                        help="Number of GPUs to use per process. This also "
                             "defines the number of parallel jobs to run.")
    parser.add_argument("--force_GPU", type=str, default="",
                        help="A list of one or more GPU IDs "
                             "(comma separated) from which GPU resources "
                             "will supplied to each split, independent of"
                             " the current memory usage of the GPUs.")
    parser.add_argument("--ignore_GPU", type=str, default="",
                        help="A list of one or more GPU IDs "
                             "(comma separated) that will not be considered.")
    parser.add_argument("--num_jobs", type=int, default=None,
                        help="OBS: Only in effect when --num_GPUs=0. Sets"
                             " the number of jobs to run in parallel when no"
                             " GPUs are attached to each job.")
    parser.add_argument("--run_on_split", type=int, default=None,
                        help="Only run a specific split")
    parser.add_argument("--script_prototype", type=str, default="./script",
                        help="Path to text file listing commands and "
                             "arguments to execute under each sub-exp folder.")
    parser.add_argument("--hparams_prototype", type=str,
                        default="./train_hparams.yaml",
                        help="Prototype hyperparameter yaml file from which"
                             " sub-CV files will be made.")
    parser.add_argument("--no_hparams", action="store_true",
                        help="Do not move a hyperparameter yaml file into "
                             "each split dir (one must be already there).")
    parser.add_argument("--start_from", type=int, default=0,
                        help="Start from CV split<start_from>. Default 0.")
    parser.add_argument("--wait_for", type=str, default="",
                        help="Waiting for PID to terminate before starting "
                             "training process.")
    parser.add_argument("--monitor_GPUs_every", type=int, default=None,
                        help="If specified, start a background process which"
                             " monitors every 'monitor_GPUs_every' seconds "
                             "whether new GPUs have become available than may"
                             " be included in the CV experiment GPU resource "
                             "pool.")
    return parser


def get_CV_folders(dir):
    key = lambda x: int(x.split("_")[-1])
    return [os.path.join(dir, p) for p in sorted(os.listdir(dir), key=key)]


def _get_GPU_sets(free_gpus, num_GPUs):
    free_gpus = list(map(str, free_gpus))
    return [",".join(free_gpus[x:x + num_GPUs]) for x in range(0, len(free_gpus),
                                                               num_GPUs)]


def get_free_GPU_sets(num_GPUs, ignore_gpus=None):
    from mpunet.utils.system import GPUMonitor
    mon = GPUMonitor()
    ignore_gpus = _gpu_string_to_list(ignore_gpus or "", as_int=True)
    free_gpus = sorted(mon.free_GPUs, key=lambda x: int(x))
    mon.stop()
    free_gpus = list(filter(lambda gpu: gpu not in ignore_gpus, free_gpus))
    total_GPUs = len(free_gpus)

    if total_GPUs % num_GPUs or not free_gpus:
        if total_GPUs < num_GPUs:
            raise ValueError("Invalid number of GPUs per process '%i' for total "
                             "GPU count of '%i' - must be evenly divisible." %
                             (num_GPUs, total_GPUs))
        else:
            raise NotImplementedError
    else:
        return _get_GPU_sets(free_gpus, num_GPUs)


def monitor_GPUs(every, gpu_queue, num_GPUs, ignore_GPU, current_pool, stop_event):
    import time
    # Make flat version of the list of gpu sets
    current_pool = [gpu for sublist in current_pool for gpu in sublist.split(",")]
    while not stop_event.is_set():
        # Get available GPU sets. Will raise ValueError if no full set is
        # available
        try:
            gpu_sets = get_free_GPU_sets(num_GPUs, ignore_GPU)
            for gpu_set in gpu_sets:
                if any([g in current_pool for g in gpu_set.split(",")]):
                    # If one or more GPUs are already in use - this may happen
                    # initially as preprocessing occurs in a process before GPU
                    # memory has been allocated - ignore the set
                    continue
                else:
                    gpu_queue.put(gpu_set)
                    current_pool += gpu_set.split(",")
        except ValueError:
            pass
        finally:
            time.sleep(every)


def parse_script(script, GPUs):
    commands = []
    with open(script) as in_file:
        for line in in_file:
            line = line.strip(" \n")
            if not line or line[0] == "#":
                continue
            # Split out in-line comments
            line = line.split("#")[0]
            # Get all arguments, remove if concerning GPU (controlled here)
            cmd = list(filter(lambda x: "gpu" not in x.lower(), line.split()))
            if "python" in line or line[:2] == "mp" or line[:2] == "ds":
                cmd.append("--force_GPU=%s" % GPUs)
            commands.append(cmd)
    return commands


def run_sub_experiment(split_dir, out_dir, script, hparams, no_hparams,
                       GPUs, GPU_queue, lock, logger):

    # Create sub-directory
    split = os.path.split(split_dir)[-1]
    out_dir = os.path.join(out_dir, split)
    out_hparams = os.path.join(out_dir, "train_hparams.yaml")
    create_folders(out_dir)

    # Get list of commands
    commands = parse_script(script, GPUs)

    # Move hparams and script files into folder
    if not no_hparams:
        copy_yaml_and_set_data_dirs(in_path=hparams,
                                    out_path=out_hparams,
                                    data_dir=split_dir)

    # Change directory and file permissions
    os.chdir(out_dir)

    # Log
    lock.acquire()
    s = "[*] Running experiment: %s" % split
    logger("\n%s\n%s" % ("-" * len(s), s))
    logger("Data dir:", split_dir)
    logger("Out dir:", out_dir)
    logger("Using GPUs:", GPUs)
    logger("\nRunning commands:")
    for i, command in enumerate(commands):
        logger(" %i) %s" % (i+1, " ".join(command)))
    logger("-"*len(s))
    lock.release()

    # Run the commands
    run_next_command = True
    for command in commands:
        if not run_next_command:
            break
        lock.acquire()
        logger("[%s - STARTING] %s" % (split, " ".join(command)))
        lock.release()
        p = subprocess.Popen(command, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        _, err = p.communicate()
        rc = p.returncode
        lock.acquire()
        if rc != 0:
            logger("[%s - ERROR - Exit code %i] %s" % (split, rc, " ".join(command)))
            logger("\n----- START error message -----\n%s\n"
                   "----- END error message -----\n" % err.decode("utf-8"))
            run_next_command = False
        else:
            logger("[%s - FINISHED] %s" % (split, " ".join(command)))
        lock.release()

    # Add the GPUs back into the queue
    GPU_queue.put(GPUs)


def _assert_run_split(start_from, monitor_GPUs_every, num_jobs):
    if start_from != 0:
        raise ValueError("Should specify either --run_on_split <split> or "
                         "--start_from <split>, got both.")
    if monitor_GPUs_every is not None:
        raise ValueError("--monitor_GPUs_every is not a valid argument"
                         " to use with --run_on_split.")
    if num_jobs != 1:
        raise ValueError("--num_jobs is not a valid argument to use with"
                         " --run_on_split.")


def _assert_force_and_ignore_gpus(force_gpu, ignore_gpu):
    force_gpu = _gpu_string_to_list(force_gpu)
    ignore_gpu = _gpu_string_to_list(ignore_gpu)
    overlap = set(force_gpu) & set(ignore_gpu)
    if overlap:
        raise RuntimeError("Cannot both force and ignore GPU(s) {}. "
                           "Got forced GPUs {} and ignored GPUs {}".format(
            overlap, force_gpu, ignore_gpu
        ))


def _gpu_string_to_list(gpu_list, as_int=False):
    str_gpus = list(filter(None, gpu_list.replace(" ", "").split(",")))
    if as_int:
        return list(map(int, str_gpus))
    return str_gpus


def entry_func(args=None):
    # Get parser
    parser = vars(get_parser().parse_args(args))

    # Get parser arguments
    cv_dir = os.path.abspath(parser["CV_dir"])
    out_dir = os.path.abspath(parser["out_dir"])
    create_folders(out_dir)
    await_PID = parser["wait_for"]
    run_split = parser["run_on_split"]
    start_from = parser["start_from"] or 0
    num_jobs = parser["num_jobs"] or 1

    # GPU settings
    num_GPUs = parser["num_GPUs"]
    force_GPU = parser["force_GPU"]
    ignore_GPU = parser["ignore_GPU"]
    monitor_GPUs_every = parser["monitor_GPUs_every"]

    # User input assertions
    _assert_force_and_ignore_gpus(force_GPU, ignore_GPU)
    if run_split:
        _assert_run_split(start_from, monitor_GPUs_every, num_jobs)

    # Wait for PID?
    if await_PID:
        from mpunet.utils import await_PIDs
        await_PIDs(await_PID)

    # Get file paths
    script = os.path.abspath(parser["script_prototype"])
    hparams = os.path.abspath(parser["hparams_prototype"])
    no_hparams = parser["no_hparams"]

    # Get list of folders of CV data to run on
    cv_folders = get_CV_folders(cv_dir)
    if run_split is not None:
        if run_split < 0 or run_split >= len(cv_folders):
            raise ValueError("--run_on_split should be in range [0-{}], "
                             "got {}".format(
                len(cv_folders)-1, run_split
            ))
        cv_folders = [cv_folders[run_split]]
        log_appendix = "_split{}".format(run_split)
    else:
        log_appendix = ""

    # Get a logger object
    logger = Logger(base_path="./", active_file="output" + log_appendix,
                    print_calling_method=False, overwrite_existing=True)

    if force_GPU:
        # Only these GPUs fill be chosen from
        from mpunet.utils import set_gpu
        set_gpu(force_GPU)
    if num_GPUs:
        # Get GPU sets (up to the number of splits)
        gpu_sets = get_free_GPU_sets(num_GPUs, ignore_GPU)[:len(cv_folders)]
    elif not num_jobs or num_jobs < 0:
        raise ValueError("Should specify a number of jobs to run in parallel "
                         "with the --num_jobs flag when using 0 GPUs pr. "
                         "process (--num_GPUs=0 was set).")
    else:
        gpu_sets = ["''"] * parser["num_jobs"]

    # Get process pool, lock and GPU queue objects
    lock = Lock()
    gpu_queue = Queue()
    for gpu in gpu_sets:
        gpu_queue.put(gpu)

    procs = []
    if monitor_GPUs_every is not None and monitor_GPUs_every:
        logger("\nOBS: Monitoring GPU pool every %i seconds\n" % monitor_GPUs_every)
        # Start a process monitoring new GPU availability over time
        stop_event = Event()
        t = Process(target=monitor_GPUs, args=(monitor_GPUs_every, gpu_queue,
                                               num_GPUs, ignore_GPU,
                                               gpu_sets, stop_event))
        t.start()
        procs.append(t)
    else:
        stop_event = None
    try:
        for cv_folder in cv_folders[start_from:]:
            gpus = gpu_queue.get()
            t = Process(target=run_sub_experiment,
                        args=(cv_folder, out_dir, script, hparams,
                              no_hparams, gpus, gpu_queue,
                              lock, logger))
            t.start()
            procs.append(t)
            for t in procs:
                if not t.is_alive():
                    t.join()
    except KeyboardInterrupt:
        for t in procs:
            t.terminate()
    if stop_event is not None:
        stop_event.set()
    for t in procs:
        t.join()


if __name__ == "__main__":
    entry_func()



from .utils import (get_free_gpu, get_best_model, get_last_model,
                    create_folders, print_options_context,
                    pred_to_class, await_and_set_free_gpu, set_gpu,
                    add_noise_to_views, DummyContext, highlighted,
                    get_free_gpus, wait_for, get_lr_at_epoch,
                    clear_csv_after_epoch, random_split, check_kwargs,
                    arr_to_fixed_precision_string, ensure_list_or_tuple,
                    await_PIDs, get_last_epoch)



import os
import numpy as np
import nibabel as nib

from argparse import ArgumentParser
from scipy.ndimage.filters import gaussian_filter
from mpunet.utils.utils import create_folders


def get_argparser():
    parser = ArgumentParser(description='Create a toy dataset - Useful '
                                        'for testing purposes')
    parser.add_argument("--out_dir", type=str, required=True,
                        help="Path to a dir in which the toy data will be "
                             "stored. Must be a non-existing or empty dir.")
    parser.add_argument("--N", type=int, default=10, required=False,
                        help='The number of train, val and test images to '
                             'create.')
    parser.add_argument("--image_size", type=int, default=128,
                        help="Voxel size of image.")
    parser.add_argument("--N_train", type=int, default=0, required=False,
                        help='The number of train images to create. '
                             'Overrides the --N flag.')
    parser.add_argument("--N_val", type=int, default=0, required=False,
                        help='The number of val images to create. '
                             'Overrides the --N flag.')
    parser.add_argument("--N_test", type=int, default=0, required=False,
                        help='The number of test images to create. '
                             'Overrides the --N flag.')
    parser.add_argument("--image_subdir", type=str, default="images",
                        help="Optional name of subdir to store image files")
    parser.add_argument("--label_subdir", type=str, default="labels",
                        help="Optional name of subdir to store label files")
    parser.add_argument("--seed", type=int, default=0,
                        help="Use a specific seed for random number "
                             "generation. Useful for debugging purposes.")
    return parser


def _elastic_deform(xx, yy, zz, sigma, alpha):
    dx = gaussian_filter((np.random.rand(*xx.shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha
    dy = gaussian_filter((np.random.rand(*yy.shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha
    dz = gaussian_filter((np.random.rand(*zz.shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha
    return xx+dx, yy+dy, zz+dz


def _get_center(img_size):
    min_, max_ = img_size * 0.05, img_size * 0.95
    range_ = max_ - min_
    center = np.random.rand(3) * range_ - min_
    return center.astype(np.int)


def create_toy_image(img_size):
    # Create empty image
    im = np.zeros(shape=[img_size]*3, dtype=np.float64)

    # Create grid
    xx, yy, zz = np.mgrid[:img_size, :img_size, :img_size]
    xx_d, yy_d, zz_d = _elastic_deform(xx, yy, zz,
                                       sigma=img_size/6,
                                       alpha=img_size*60)

    # Add background
    im += np.sin(0.05 + np.random.rand() * xx/img_size) + \
          np.power(np.cos(0.05 + np.random.rand() * yy/img_size), 0.5) + \
          np.power(np.sin(0.05 + np.random.rand() * zz/img_size), 2)

    # Scale 0-0.75
    im -= im.min()
    im /= im.max() * (1/0.75)

    # Add sphere
    radius = img_size/8 + np.random.rand()*img_size/2
    center = _get_center(img_size)
    sphere_mask = np.power(xx_d-center[0], 2) + \
                  np.power(yy_d-center[1], 2) + \
                  np.power(zz_d-center[2], 2) <= np.power(radius, 2)
    sphere_vals = np.sin((xx[sphere_mask]-center[0])/img_size) + \
                  np.sin((yy[sphere_mask]-center[1])/img_size*5) + \
                  np.sin((zz[sphere_mask]-center[2])/img_size*10)

    # Add and scale 0.2-1.0
    im[sphere_mask] += sphere_vals
    im[sphere_mask] -= im[sphere_mask].min()
    im[sphere_mask] /= im[sphere_mask].max() * 1/0.8
    im[sphere_mask] += 0.2

    # Add elastic square
    size = img_size/4 + np.random.rand()*img_size/2
    center = _get_center(img_size)
    square_mask = (np.abs(xx_d-center[0]) < size/2) & \
                  (np.abs(yy_d-center[1]) < size/2) & \
                  (np.abs(zz_d-center[2]) < size/2)
    square_vals = np.power(xx[square_mask]-center[0], 2) + \
                  np.power(yy[square_mask]-center[1], 2) + \
                  np.power(zz[square_mask]-center[2], 2)

    # Add and scale 0.2-1.0
    im[square_mask] += square_vals
    im[square_mask] += square_vals
    im[square_mask] -= im[square_mask].min()
    im[square_mask] /= im[square_mask].max() * 1/0.8
    im[square_mask] += 0.2

    # Add torus
    center_radius = img_size/16 + np.random.rand()*img_size/6
    tube_radius = img_size/32 + np.random.rand()*img_size/12
    center = _get_center(img_size)
    torus_mask_1 = np.power(center_radius-np.sqrt(np.power(xx_d-center[0], 2) +
                                                  np.power(yy_d-center[1], 2)), 2) + \
                   np.power(zz_d-center[2], 2) <= np.power(tube_radius, 2)
    torus_mask_2 = np.power(center_radius-np.sqrt(np.power(zz_d-center[2], 2) +
                                                  np.power(yy_d-center[1], 2)), 2) + \
                   np.power(xx_d-center[0], 2) <= np.power(tube_radius, 2)
    torus_mask = np.logical_or(torus_mask_1, torus_mask_2)
    torus_vals = np.random.poisson(im[torus_mask])

    # Add and scale 0.2-1.0
    im[torus_mask] += torus_vals
    im[torus_mask] -= im[torus_mask].min()
    im[torus_mask] /= im[torus_mask].max() * 1 / 0.8
    im[torus_mask] += 0.2

    # Add noise to image
    im += np.random.poisson(lam=10, size=im.shape) * np.random.randn(*im.shape) * 0.0005

    # Scale overall image to [0, ..., 1]
    im -= im.min()
    im /= im.max()

    # Scale to 16 bit int
    im *= np.iinfo(np.uint16).max
    return im.astype(np.uint16), sphere_mask, square_mask, torus_mask


def create_toy_labels(*masks):
    assert len(masks) < np.iinfo(np.uint8).max
    labels = np.zeros(shape=masks[0].shape, dtype=np.uint8)
    for i, mask in enumerate(masks):
        labels[mask] = i+1
    return labels


def create_toy_data_point(img_size):
    im, m1, m2, m3 = create_toy_image(img_size)
    lab = create_toy_labels(m1, m2, m3)
    return im, lab


def create_toy_dataset(N, image_size, out_dir, image_subdir, label_subdir):
    image_out_dir = os.path.join(out_dir, image_subdir)
    label_out_dir = os.path.join(out_dir, label_subdir)
    create_folders([out_dir, image_out_dir, label_out_dir])

    for img_id in range(N):
        print("--", img_id)
        file_name = "toy_data_{}.nii.gz".format(img_id)
        im_path = os.path.join(image_out_dir, file_name)
        lab_path = os.path.join(label_out_dir, file_name)
        image, label = create_toy_data_point(image_size)

        affine = np.eye(4)
        nib.save(nib.Nifti1Image(image, affine=affine), im_path)
        nib.save(nib.Nifti1Image(label, affine=affine), lab_path)


def entry_func(args=None):

    args = get_argparser().parse_args(args)
    out_dir = os.path.abspath(os.path.abspath(args.out_dir))
    if os.path.exists(out_dir) and os.listdir(out_dir):
        from sys import exit
        print("Path '{}' already exists and is not empty!".format(args.out_dir))
        exit(1)
    create_folders([out_dir], create_deep=False)

    n_train, n_val, n_test = [args.N] * 3
    n_train = args.N_train or n_train
    n_val = args.N_val or n_val
    n_test = args.N_test or n_test

    np.random.seed(args.seed)
    for dataset, N in zip(["train", "val", "test"], [n_train, n_val, n_test]):
        if not N:
            continue
        else:
            print("[*] Creating dataset '{}' of {} samples".format(
                dataset, N
            ))
            create_toy_dataset(N=N,
                               image_size=args.image_size,
                               out_dir=os.path.join(out_dir, dataset),
                               image_subdir=args.image_subdir,
                               label_subdir=args.label_subdir)


if __name__ == "__main__":
    entry_func()



from mpunet.logging.default_logger import ScreenLogger
from mpunet.image.queue import LazyQueue, BaseQueue


def get_sequence(data_queue,
                 is_validation,
                 logger=None,
                 augmenters=None,
                 **seq_kwargs):
    """
    Prepares a MultiPlanar.sequence object for generating batches of data from
    a set of images contained in a mpunet.image.queue typed object.

    These generator-like objects pull data from ImagePairs during
    training as needed. The sequences are specific to the model type (for
    instance 2D and 3D models have separate sequence classes) and may
    differ in the interpolation schema as well (voxel vs. iso scanner space
    coordinates for instance).

    Args:
        data_queue:    mpunet.image.queue data queue type object
        is_validation: Boolean, is this a validation sequence? (otherwise
                       training)
        logger:        TODO
        augmenters:    TODO
        **seq_kwargs:  Additional arguments passed to the Sequencer

    Raises:
        ValueError if intrp_style is not valid
    """
    logger = logger or ScreenLogger()
    # If data_queue in ImagePairDataset
    if not isinstance(data_queue, BaseQueue):
        # Wrap the passed ImagePairLoader in a LazyQueue
        data_queue = LazyQueue(data_queue)
    aug_list = []
    if not is_validation:
        # On the fly augmentation?
        list_of_aug_dicts = augmenters
        if list_of_aug_dicts:
            logger("Using on-the-fly augmenters:")
            from mpunet.augmentation import augmenters
            for aug in list_of_aug_dicts:
                aug_cls = augmenters.__dict__[aug["cls_name"]](**aug["kwargs"])
                aug_list.append(aug_cls)
                logger(aug_cls)

    if seq_kwargs['intrp_style'].lower() == "iso_live":
        # Isotrophic 2D plane sampling
        from mpunet.sequences import IsotrophicLiveViewSequence2D
        return IsotrophicLiveViewSequence2D(data_queue,
                                            is_validation=is_validation,
                                            list_of_augmenters=aug_list,
                                            logger=logger,
                                            **seq_kwargs)
    elif seq_kwargs['intrp_style'].lower() == "iso_live_3d":
        # Isotrophic 3D box sampling
        from mpunet.sequences import IsotrophicLiveViewSequence3D
        return IsotrophicLiveViewSequence3D(data_queue,
                                            is_validation=is_validation,
                                            list_of_augmenters=aug_list,
                                            logger=logger,
                                            **seq_kwargs)
    elif seq_kwargs['intrp_style'].lower() == "patches_3d":
        # Random selection of boxes
        from mpunet.sequences import PatchSequence3D
        return PatchSequence3D(data_queue,
                               is_validation=is_validation,
                               list_of_augmenters=aug_list, **seq_kwargs)
    elif seq_kwargs['intrp_style'].lower() == "sliding_patches_3d":
        # Sliding window selection of boxes
        from mpunet.sequences import SlidingPatchSequence3D
        return SlidingPatchSequence3D(data_queue,
                                      is_validation=is_validation,
                                      list_of_augmenters=aug_list,
                                      **seq_kwargs)
    else:
        raise ValueError("Invalid interpolator schema '%s' specified"
                         % seq_kwargs['intrp_style'])





class ScreenLogger(object):
    """
    Minimal wrapper class around the built-in print function replicating some
    functionality of the mpunet Logger class so that this class can be
    used in a similar fashion while only printing to screen and not a log file
    """
    def __init__(self, print_to_screen=True):
        self.print_to_screen = print_to_screen

    def __call__(self,
                 *args,
                 print_to_screen=None,
                 print_calling_method=None,
                 no_print=None,
                 **kwargs):
        if self.print_to_screen and not no_print:
            print(*args, **kwargs)

    def warn(self, *args, **kwargs):
        self.__call__("[WARNING]", *args, **kwargs)

    def __enter__(self):
        return

    def __exit__(self, *args):
        return



from tensorflow.keras.utils import Sequence
from mpunet.logging import ScreenLogger


class MultiTaskSequence(Sequence):
    """
    A Sequence which simply wraps around multiple other Sequences, calling each
    of them to create batches for each Sequence (typically different 'tasks')
    and returns three lists, stroing iamges, labels and weights, each of
    length equal to the number of wrapped Sequencers.
    """
    def __init__(self, sequencers, task_names, logger=None):
        super().__init__()
        self.logger = logger or ScreenLogger()
        self.task_names = task_names
        self.sequencers = sequencers
        self.log()

        # Redirect setattrs to the sub-sequences
        self.redirect = True

    def log(self):
        self.logger("--- MultiTaskSequence sequencer --- ",
                    print_calling_method=True)
        self.logger("N tasks:  %i" % len(self.sequencers))

    def __iter__(self):
        for s in self.sequencers:
            yield s

    @property
    def n_samples(self):
        return sum([s.n_samples for s in self])

    def __getattr__(self, item):
        # Fetch item from all sub-sequences
        return [getattr(s, item) for s in self.sequencers]

    def __setattr__(self, key, value):
        if self.__dict__.get("redirect", False):
            # Fetch item from all sub-sequences
            for s in self:
                # Set on all sub-sequences
                setattr(s, key, value)
        else:
            # Set on self
            self.__dict__[key] = value

    def __len__(self):
        """Number of batch in the Sequence.

        Returns:
            The number of batches in the Sequence.
        """
        return sum([len(seq) for seq in self.sequencers])

    def __getitem__(self, index):
        batches_x, batches_y, batches_w = [], [], []
        for seq in self.sequencers:
            x, y, w = seq[0]
            batches_x.append(x)
            batches_y.append(y)
            batches_w.append(w)

        return batches_x, batches_y, batches_w



import time
import os

from multiprocessing import Process, Event, Queue
from mpunet.utils.utils import get_free_gpus, _get_free_gpu, set_gpu
from mpunet.logging import ScreenLogger


class GPUMonitor(Process):
    def __init__(self):
        self.logger = ScreenLogger()

        # Prepare signal
        self.stop_signal = Event()
        self.run_signal = Event()
        self.set_signal = Event()

        # Stores list of available GPUs
        self._free_GPUs = Queue()

        super(GPUMonitor, self).__init__(target=self._monitor)
        self.start()

    def stop(self):
        self.stop_signal.set()

    def _monitor(self):
        while not self.stop_signal.is_set():
            if self.run_signal.is_set():
                # Empty queue, just in case...?
                self._free_GPUs.empty()

                # Get free GPUs
                free = get_free_gpus()

                # Add number of elements that will be put in the queue as first
                # element to be read from main process
                self._free_GPUs.put(len(free))

                # Add available GPUs to queue
                for g in free:
                    self._free_GPUs.put(g)

                # Set the signal that main process can start reading queue
                self.set_signal.set()

                # Stop run signal for this process
                self.run_signal.clear()
            else:
                time.sleep(0.5)
                self.set_signal.clear()

    @property
    def free_GPUs(self):
        self.run_signal.set()
        while not self.set_signal.is_set():
            time.sleep(0.2)

        free = []
        for i in range(self._free_GPUs.get()):
            free.append(self._free_GPUs.get())
        return free

    def get_free_GPUs(self, N):
        return _get_free_gpu(self.free_GPUs, N)

    def await_and_set_free_GPU(self, N=0, sleep_seconds=60, stop_after=False):
        cuda_visible_dev = ""
        if N != 0:
            self.logger("Waiting for free GPU.")
            found_gpu = False
            while not found_gpu:
                cuda_visible_dev = self.get_free_GPUs(N=N)
                if cuda_visible_dev:
                    self.logger("Found free GPU: %s" % cuda_visible_dev)
                    found_gpu = True
                else:
                    self.logger("No available GPUs... Sleeping %i seconds." % sleep_seconds)
                    time.sleep(sleep_seconds)
        else:
            self.logger("Using CPU based computations only!")
        self.set_GPUs = cuda_visible_dev
        if stop_after:
            self.stop()

    @property
    def num_currently_visible(self):
        return len(self.set_GPUs.strip().split(","))

    @property
    def set_GPUs(self):
        try:
            return os.environ["CUDA_VISIBLE_DEVICES"]
        except KeyError:
            return ""

    @set_GPUs.setter
    def set_GPUs(self, GPUs):
        set_gpu(GPUs)

    def set_and_stop(self, GPUs):
        self.set_GPUs = GPUs
        self.stop()



import unittest
import os
import numpy as np


class TestImagePairWithValidImage(unittest.TestCase):
    """
    Tests basic properties and error raising functionalities of an ImagePair
    object initialized with a (valid) nii.gz file stored on disk.
    """
    # Defines the properties of the (valid) image
    image_path = ".temp_valid_image.nii.gz"
    data = np.random.randn(12, 14, 16, 3).astype(np.float64)
    affine = np.diag([1, 0.5, 0.1, 1])

    @classmethod
    def setUpClass(cls):
        """ Save a temperature, valid image (nii.gz) file to disk """
        if os.path.exists(cls.image_path):
            raise OSError("Out path {} already exists".format(cls.image_path))
        import nibabel as nib
        nii = nib.Nifti1Image(cls.data, affine=cls.affine)
        nib.save(nii, cls.image_path)

    @classmethod
    def tearDownClass(cls):
        """ Remove the temporary nii.gz file """
        if os.path.exists(cls.image_path):
            os.remove(cls.image_path)

    def setUp(self):
        """ Load an ImagePair object from the .nii.gz file stored on disk """
        from mpunet.image.image_pair import ImagePair
        self.im = ImagePair(img_path=self.image_path)

    def test_stored_image_matches_disk_image(self):
        """ Tests if the stored image matches the Nifti image saved to disk """

        # Assert some fields are as expected
        self.assertTrue(self.im.predict_mode,
                        msg="ImagePair initialized with only an image should "
                            "have attribute predict_mode=True.")
        self.assertListEqual(list(self.im.shape), list(self.data.shape),
                             msg="The dimensionality of the loaded image "
                                 "object does not match the expected "
                                 "(got {}, expected {})".format(
                                 self.im.image.shape, self.data.shape))
        self.assertTrue(np.isclose(self.im.image, self.data).all(),
                        msg="The data of the loaded image is not (nearly) "
                            "identical to the data that was saved to disk.")
        self.assertTrue(self.im.image.dtype == np.dtype("float32"),
                        msg="The stored image has dtype {}, but ImagePair "
                            "should always store float32 "
                            "images".format(self.im.image.dtype))
        self.assertTrue(np.isclose(self.im.affine, self.affine).all(),
                        msg="The stored affine matrix is not (nearly) "
                            "identical to the one saved to disk.")

    def test_error_raising(self):
        """ Asserts that proper errors are raised with illegal usage """

        # Check errors are raised with illegal actions
        from mpunet.errors.image_errors import (NoLabelFileError,
                                                ReadOnlyAttributeError)
        with self.assertRaises(NoLabelFileError,
                               msg="Referencing the labels attribute of an "
                                   "ImagePair that has no labels should raise "
                                   "a {}".format(NoLabelFileError.__name__)):
            _ = self.im.labels
        with self.assertRaises(ReadOnlyAttributeError,
                               msg="Trying to set the 'image' attribute on an "
                                   "ImagePair object should raise a "
                                   "{}".format(ReadOnlyAttributeError.__name__)):
            self.im.image = [1, 2, 3]
        with self.assertRaises(ReadOnlyAttributeError,
                               msg="Trying to set the 'labels' attribute on an"
                                   " ImagePair object should raise a "
                                   "{}".format(ReadOnlyAttributeError.__name__)):
            self.im.labels = [1, 2, 3]
        with self.assertRaises(ReadOnlyAttributeError,
                               msg="Trying to set the 'affine' attribute on an"
                                   " ImagePair object should raise a "
                                   "{}".format(ReadOnlyAttributeError.__name__)):
            self.im.affine = [1, 2, 3]

    def test_shape_values(self):
        """ Tests that the ImagePair stores correct image coordinates """
        
        # Check voxel coordinates center
        self.assertListEqual(list(self.im.center), [5.5, 6.5, 7.5],
                             msg="Error in calculation of image voxel center "
                                 "(zero-indexed) coordinates")

        # Check scanner space center
        expected = np.array([5.5, 3.25, 0.75])
        got = np.array(self.im.real_center)
        self.assertTrue(np.isclose(got, expected).all(),
                        msg="Error in calculation of image scanner "
                            "space center coordinates. Expected "
                            "{}, got {}".format(expected, got))

        # Check real shape
        expected = np.array([12, 7, 1.6])
        got = np.array(self.im.real_shape)
        self.assertTrue(np.isclose(got, expected).all(),
                        msg="Error in calculation of image real shape "
                            "(mm in scanner space). Expected {}, "
                            "got {}".format(expected, got))



from setuptools import setup, find_packages

from mpunet import __version__

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

with open("requirements.txt") as req_file:
    requirements = list(filter(None, req_file.read().split("\n")))


setup(
    name='mpunet',
    version=__version__,
    description='Multi-Planar UNet for autonomous '
                'segmentation of 3D medical images',
    long_description=readme + "\n\n" + history,
    long_description_content_type='text/markdown',
    author='Mathias Perslev',
    author_email='map@di.ku.dk',
    url='https://github.com/perslev/MultiPlanarUNet',
    license='LICENSE.txt',
    packages=find_packages(),
    package_dir={'mpunet':
                 'mpunet'},
    include_package_data=True,
    setup_requires=["setuptools_git>=0.3",],
    entry_points={
       'console_scripts': [
           'mp=mpunet.bin.mp:entry_func',
       ],
    },
    install_requires=requirements,
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'Operating System :: POSIX',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'License :: OSI Approved :: MIT License']
)



from .unet import UNet
from .unet3D import UNet3D
from .fusion_model import FusionModel
from .model_init import model_initializer
from .multitask_unet2d import MultiTaskUNet2D

# Prepare a dictionary mapping from model names to data prep. functions
from mpunet.preprocessing import data_preparation_funcs as dpf

PREPARATION_FUNCS = {
    "UNet": dpf.prepare_for_multi_view_unet,
    "UNet3D": dpf.prepare_for_3d_unet,
    "MultiTaskUNet2D": dpf.prepare_for_multi_task_2d
}



from .metrics import dice_all
from .loss_functions import (SparseExponentialLogarithmicLoss,
                             SparseDiceLoss,
                             SparseExpLogDice,
                             SparseFocalLoss,
                             SparseGeneralizedDiceLoss,
                             SparseJaccardDistanceLoss)



import argparse
import os

"""
Entry script redirecting all command line arguments to a specified script
from the mpunet.bin folder.

Usage:
mp [script] [script args...]
"""


def get_parser():
    from mpunet import bin, __version__
    import pkgutil
    mods = pkgutil.iter_modules(bin.__path__)

    ids = "Multi-Planar UNet ({})".format(__version__)
    sep = "-" * len(ids)
    usage = ("mp [script] [script args...]\n\n"
             "%s\n%s\n"
             "Available scripts:\n") % (ids, sep)

    choices = []
    file_name = os.path.split(os.path.abspath(__file__))[-1]
    for m in mods:
        if isinstance(m, tuple):
            name, ispkg = m[1], m[2]
        else:
            name, ispkg = m.name, m.ispkg
        if name == file_name[:-3] or ispkg:
            continue
        usage += "- " + name + "\n"
        choices.append(name)

    # Top level parser
    parser = argparse.ArgumentParser(usage=usage)
    parser.add_argument("script", help="Name of the mp script to run.",
                        choices=choices)
    parser.add_argument("args", help="Arguments passed to script",
                        nargs=argparse.REMAINDER)
    return parser


def entry_func(args=None):
    # Get the script to execute, parse only first input
    parsed = get_parser().parse_args(args)
    script = parsed.script

    # Import the script
    import importlib
    mod = importlib.import_module("mpunet.bin." + script)

    # Call entry function with remaining arguments
    mod.entry_func(parsed.args)


if __name__ == "__main__":
    entry_func()



"""
Mathias Perslev
MSc Bioinformatics

University of Copenhagen
November 2017
"""

from mpunet.logging import ScreenLogger
from mpunet.utils.conv_arithmetics import compute_receptive_fields

from tensorflow.keras.models import Model
from tensorflow.keras import regularizers
from tensorflow.keras.layers import (Input, BatchNormalization, Cropping3D,
                                     Concatenate, Conv3D, MaxPooling3D,
                                     UpSampling3D, Reshape)
import numpy as np


class UNet3D(Model):
    """
    3D UNet implementation with batch normalization and complexity factor adj.

    See original paper at http://arxiv.org/abs/1505.04597
    """
    def __init__(self,
                 n_classes,
                 dim=None,
                 n_channels=1,
                 depth=3,
                 out_activation="softmax",
                 activation="relu",
                 kernel_size=3,
                 padding="same",
                 complexity_factor=1,
                 flatten_output=False,
                 l2_reg=None,
                 logger=None, **kwargs):
        """
        n_classes (int):
            The number of classes to model, gives the number of filters in the
            final 1x1 conv layer.
        dim (int):
            Box dimensionality (on all three axes)
            Note that depending on image dims cropping may
            be necessary. To avoid this, use image dimensions DxDxD for which
            D * (1/2)^n is an integer, where n is the number of (2x2)
            max-pooling layers; in this implementation 4.
            For n=4, D \in {..., 192, 208, 224, 240, 256, ...} etc.
        n_channels (int):
            Number of channels in the input image.
        depth (int):
            Number of conv blocks in encoding layer (number of 2x2x2 max pools)
            Note: each block doubles the filter count while halving the spatial
            dimensions of the features.
        out_activation (string):
            Activation function of output 1x1x1 conv layer. Usually one of
            'softmax', 'sigmoid' or 'linear'.
        activation (string):
            Activation function for convolution layers
        kernel_size (int):
            Kernel size for convolution layers
        padding (string):
            Padding type ('same' or 'valid')
        complexity_factor (int/float):
            Use int(N * sqrt(complexity_factor)) number of filters in each
            3D convolution layer instead of default N.
        l2_reg (float in [0, 1])
            L2 regularization on Conv3D weights
        logger (mpunet.logging.Logger | ScreenLogger):
            MutliViewUNet.Logger object, logging to files or screen.
        """
        # Set logger or standard print wrapper
        self.logger = logger or ScreenLogger()

        # Set various attributes
        self.img_shape = (dim, dim, dim, n_channels)
        self.n_classes = n_classes
        self.cf = np.sqrt(complexity_factor)
        self.kernel_size = kernel_size
        self.activation = activation
        self.out_activation = out_activation
        self.l2_reg = l2_reg
        self.padding = padding
        self.depth = depth
        self.flatten_output = flatten_output

        # Shows the number of pixels cropped of the input image to the output
        self.label_crop = np.array([[0, 0], [0, 0], [0, 0]])

        # Build model and init base keras Model class
        super().__init__(*self.init_model())

        # Compute receptive field
        names = [x.__class__.__name__ for x in self.layers]
        index = names.index("UpSampling3D")
        self.receptive_field = compute_receptive_fields(self.layers[:index])[-1][-1]

        # Log the model definition
        self.log()

    def init_model(self):
        """
        Build the UNet model with the specified input image shape.
        """
        inputs = Input(shape=self.img_shape)

        # Apply regularization if not None or 0
        kr = regularizers.l2(self.l2_reg) if self.l2_reg else None

        """
        Encoding path
        """
        filters = 64
        in_ = inputs
        residual_connections = []
        for i in range(self.depth):
            conv = Conv3D(int(filters*self.cf), self.kernel_size,
                          activation=self.activation, padding=self.padding,
                          kernel_regularizer=kr)(in_)
            conv = Conv3D(int(filters * self.cf), self.kernel_size,
                          activation=self.activation, padding=self.padding,
                          kernel_regularizer=kr)(conv)
            bn = BatchNormalization()(conv)
            in_ = MaxPooling3D(pool_size=(2, 2, 2))(bn)

            # Update filter count and add bn layer to list for residual conn.
            filters *= 2
            residual_connections.append(bn)

        """
        Bottom (no max-pool)
        """
        conv = Conv3D(int(filters * self.cf), self.kernel_size,
                      activation=self.activation, padding=self.padding,
                      kernel_regularizer=kr)(in_)
        conv = Conv3D(int(filters * self.cf), self.kernel_size,
                      activation=self.activation, padding=self.padding,
                      kernel_regularizer=kr)(conv)
        bn = BatchNormalization()(conv)

        """
        Up-sampling
        """
        residual_connections = residual_connections[::-1]
        for i in range(self.depth):
            # Reduce filter count
            filters /= 2

            # Up-sampling block
            # Note: 2x2 filters used for backward comp, but you probably
            # want to use 3x3 here instead.
            up = UpSampling3D(size=(2, 2, 2))(bn)
            conv = Conv3D(int(filters * self.cf), 2, activation=self.activation,
                          padding=self.padding, kernel_regularizer=kr)(up)
            bn = BatchNormalization()(conv)

            # Crop and concatenate
            cropped_res = self.crop_nodes_to_match(residual_connections[i], bn)
            merge = Concatenate(axis=-1)([cropped_res, bn])

            conv = Conv3D(int(filters * self.cf), self.kernel_size,
                          activation=self.activation, padding=self.padding,
                          kernel_regularizer=kr)(merge)
            conv = Conv3D(int(filters * self.cf), self.kernel_size,
                          activation=self.activation, padding=self.padding,
                          kernel_regularizer=kr)(conv)
            bn = BatchNormalization()(conv)

        """
        Output modeling layer
        """
        out = Conv3D(self.n_classes, 1, activation=self.out_activation)(bn)
        if self.flatten_output:
            out = Reshape([np.prod(self.img_shape[:3]),
                           self.n_classes], name='flatten_output')(out)

        return [inputs], [out]

    def crop_nodes_to_match(self, node1, node2):
        """
        If necessary, applies Cropping3D layer to node1 to match shape of node2
        """
        s1 = np.array(node1.get_shape().as_list())[1:-1]
        s2 = np.array(node2.get_shape().as_list())[1:-1]

        if np.any(s1 != s2):
            c = (s1 - s2).astype(np.int)
            cr = np.array([c//2, c//2]).T
            cr[:, 1] += c % 2
            cropped_node1 = Cropping3D(cr)(node1)
            self.label_crop += cr
        else:
            cropped_node1 = node1

        return cropped_node1

    def log(self):
        self.logger("UNet3D Model Summary\n--------------------")
        self.logger("Image dims:        (%i, %i, %i)" % tuple(self.img_shape[:-1]))
        self.logger("Image channels:    %i" % self.img_shape[-1])
        self.logger("N classes:         %i" % self.n_classes)
        self.logger("CF factor:         %.3f" % self.cf**2)
        self.logger("Depth:             %i" % self.depth)
        self.logger("l2 reg:            %s" % self.l2_reg)
        self.logger("Padding:           %s" % self.padding)
        self.logger("Conv activation:   %s" % self.activation)
        self.logger("Out activation:    %s" % self.out_activation)
        self.logger("Receptive field:   %s" % self.receptive_field)
        self.logger("N params:          %i" % self.count_params())
        self.logger("Output:            %s" % self.output)
        self.logger("Crop:              %s" % (self.label_crop if np.sum(self.label_crop) != 0 else "None"))



"""
Prediction/evaluation script

Mathias Perslev
March 2018
"""

import os
from argparse import ArgumentParser


def get_argparser():
    parser = ArgumentParser(description='Predict using a mpunet model.')
    parser.add_argument("--project_dir", type=str, default="./",
                        help='Path to mpunet project folder')
    parser.add_argument("-f", help="Predict on a single file")
    parser.add_argument("-l", help="Optional single label file to use with -f")
    parser.add_argument("--data_dir", type=str, default=None,
                        help="Directory storing data. "
                             "Must contain sub-folder 'images'")
    parser.add_argument("--out_dir", type=str, default="predictions",
                        help="Output folder to store results")
    parser.add_argument("--num_GPUs", type=int, default=1,
                        help="Number of GPUs to use for this job")
    parser.add_argument("--overwrite", action='store_true',
                        help='Overwrite previous results at the output folder')
    parser.add_argument("--no_eval", action="store_true",
                        help="Perform no evaluation of the prediction performance. "
                             "No label files loaded when this flag applies.")
    parser.add_argument("--strides", type=int, default=None,
                        help="Predict on strided overlapping boxes "
                             "instead of only non-overlapping.")
    parser.add_argument("--extra", default="2x",
                        help="Sample N extra patches to perform majority voting")
    parser.add_argument("--force_GPU", type=int, default=-1)
    parser.add_argument("--save_only_pred", action="store_true",
                        help="Save only the predicted volume as .nii files ("
                             "do not save image and labels)")
    return parser


def validate_folders(base_dir, data_dir, out_dir, overwrite):

    # Check base (model) dir contains required files
    must_exist = ("train_hparams.yaml", "model")
    for p in must_exist:
        p = os.path.join(base_dir, p)
        if not os.path.exists(p):
            from sys import exit
            print("[*] Invalid mpunet project folder: '%s'"
                  "\n    Needed file/folder '%s' not found." % (base_dir, p))
            exit(0)

    # Check if output folder already exists
    if not overwrite and os.path.exists(out_dir):
        from sys import exit
        print("[*] Output directory already exists at: '%s'"
              "\n  Use --overwrite to overwrite" % out_dir)
        exit(0)
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)


def entry_func(args=None):

    # Get command line arguments
    args = vars(get_argparser().parse_args(args))
    base_dir = os.path.abspath(args["project_dir"])
    _file = args["f"]
    label = args["l"]
    N_extra = args["extra"]
    try:
        N_extra = int(N_extra)
    except ValueError:
        pass

    # Get settings from YAML file
    from mpunet.hyperparameters import YAMLHParams
    hparams = YAMLHParams(os.path.join(base_dir, "train_hparams.yaml"))

    # Set strides
    hparams["fit"]["strides"] = args["strides"]

    if not _file:
        try:
            # Data specified from command line?
            data_dir = os.path.abspath(args["data_dir"])

            # Set with default sub dirs
            hparams["test_data"] = {"base_dir": 
                                        data_dir, "img_subdir": "images",
                                    "label_subdir": "labels"}
        except (AttributeError, TypeError):
            data_dir = hparams["test_data"]["base_dir"]
    else:
        data_dir = False
    out_dir = os.path.abspath(args["out_dir"])
    overwrite = args["overwrite"]
    predict_mode = args["no_eval"]
    save_only_pred = args["save_only_pred"]

    # Check if valid dir structures
    validate_folders(base_dir, data_dir, out_dir, overwrite)

    # Import all needed modules (folder is valid at this point)
    import numpy as np
    from mpunet.image import ImagePairLoader, ImagePair
    from mpunet.utils import get_best_model, create_folders, \
                                    pred_to_class, await_and_set_free_gpu, set_gpu
    from mpunet.utils.fusion import predict_3D_patches, predict_3D_patches_binary, pred_3D_iso
    from mpunet.logging import init_result_dict_3D, save_all_3D
    from mpunet.evaluate import dice_all
    from mpunet.bin.predict import save_nii_files

    # Fetch GPU(s)
    num_GPUs = args["num_GPUs"]
    force_gpu = args["force_GPU"]
    # Wait for free GPU
    if force_gpu == -1:
        await_and_set_free_gpu(N=num_GPUs, sleep_seconds=240)
    else:
        set_gpu(force_gpu)

    # Read settings from the project hyperparameter file
    dim = hparams["build"]["dim"]
    n_classes = hparams["build"]["n_classes"]
    mode = hparams["fit"]["intrp_style"]

    # Set ImagePairLoader object
    if not _file:
        image_pair_loader = ImagePairLoader(predict_mode=predict_mode, **hparams["test_data"])
    else:
        predict_mode = not bool(label)
        image_pair_loader = ImagePairLoader(predict_mode=predict_mode,
                                            initialize_empty=True)
        image_pair_loader.add_image(ImagePair(_file, label))
    all_images = {image.identifier: image for image in image_pair_loader.images}

    # Set scaler and bg values
    image_pair_loader.set_scaler_and_bg_values(
        bg_value=hparams.get_from_anywhere('bg_value'),
        scaler=hparams.get_from_anywhere('scaler'),
        compute_now=False
    )

    # Init LazyQueue and get its sequencer
    from mpunet.sequences.utils import get_sequence
    seq = get_sequence(data_queue=image_pair_loader,
                       is_validation=True,
                       **hparams["fit"], **hparams["build"])

    """ Define UNet model """
    from mpunet.models import model_initializer
    hparams["build"]["batch_size"] = 1
    unet = model_initializer(hparams, False, base_dir)
    model_path = get_best_model(base_dir + "/model")
    unet.load_weights(model_path)

    # Evaluate?
    if not predict_mode:
        # Prepare dictionary to store results in pd df
        results, detailed_res = init_result_dict_3D(all_images, n_classes)

        # Save to check correct format
        save_all_3D(results, detailed_res, out_dir)

    # Define result paths
    nii_res_dir = os.path.join(out_dir, "nii_files")
    create_folders(nii_res_dir)

    image_ids = sorted(all_images)
    for n_image, image_id in enumerate(image_ids):
        print("\n[*] Running on: %s" % image_id)

        with seq.image_pair_queue.get_image_by_id(image_id) as image_pair:
            if mode.lower() == "iso_live_3d":
                pred = pred_3D_iso(model=unet,
                                   sequence=seq,
                                   image=image_pair,
                                   extra_boxes=N_extra,
                                   min_coverage=None)
            else:
                # Predict on volume using model
                if n_classes > 1:
                    pred = predict_3D_patches(model=unet,
                                              patches=seq,
                                              image=image_pair,
                                              N_extra=N_extra)
                else:
                    pred = predict_3D_patches_binary(model=unet,
                                                     patches=seq,
                                                     image_id=image_id,
                                                     N_extra=N_extra)

            if not predict_mode:
                # Get patches for the current image
                y = image_pair.labels

                # Calculate dice score
                print("Mean dice: ", end="", flush=True)
                p = pred_to_class(pred, img_dims=3, has_batch_dim=False)
                dices = dice_all(y, p, n_classes=n_classes, ignore_zero=True)
                mean_dice = dices[~np.isnan(dices)].mean()
                print("Dices: ", dices)
                print("%s (n=%i)" % (mean_dice, len(dices)))

                # Add to results
                results[image_id] = [mean_dice]
                detailed_res[image_id] = dices

                # Overwrite with so-far results
                save_all_3D(results, detailed_res, out_dir)

                # Save results
                save_nii_files(p, image_pair, nii_res_dir, save_only_pred)

    if not predict_mode:
        # Write final results
        save_all_3D(results, detailed_res, out_dir)


if __name__ == "__main__":
    entry_func()
